@@ -30,8 +30,8 @@
 #include "dispatcher.h"
 #include "server.h"
 #include "router.h"
-#include "aggregator.h"
 #include "fnv1a.h"
+#include "aggregator.h"
 
 static pthread_t aggregatorid;
 static size_t prevreceived = 0;
@@ -89,43 +89,17 @@ char
 aggregator_add_compute(
 		aggregator *s,
 		const char *metric,
-		const char *type)
+		enum _aggr_compute_type act,
+		unsigned char pctl)
 {
 	struct _aggr_computes *ac = s->computes;
-	enum _aggr_compute_type act;
 	char store = 0;
-	int pctl = 0;
 
-	if (strcmp(type, "sum") == 0) {
-		act = SUM;
-	} else if (strcmp(type, "count") == 0 || strcmp(type, "cnt") == 0) {
-		act = CNT;
-	} else if (strcmp(type, "max") == 0) {
-		act = MAX;
-	} else if (strcmp(type, "min") == 0) {
-		act = MIN;
-	} else if (strcmp(type, "average") == 0 || strcmp(type, "avg") == 0) {
-		act = AVG;
-	} else if (strcmp(type, "median") == 0) {
-		act = MEDN;
+	if (act == MEDN) {
 		pctl = 50;
 		store = 1;
-	} else if (strncmp(type, "percentile", strlen("percentile")) == 0) {
-		pctl = atoi(type + strlen("percentile"));
-		if (pctl > 100 || pctl <= 0) {
-			return -1;
-		} else {
-			act = PCTL;
-			store = 1;
-		}
-	} else if (strcmp(type, "variance") == 0) {
-		act = VAR;
+	} else if (act == PCTL || act == VAR || act == SDEV) {
 		store = 1;
-	} else if (strcmp(type, "stddev") == 0) {
-		act = SDEV;
-		store = 1;
-	} else {
-		return -1;
 	}
 
 	if (ac == NULL) {
@@ -135,10 +109,14 @@ aggregator_add_compute(
 			ac = ac->next;
 		ac = ac->next = malloc(sizeof(*ac));
 	}
+	if (ac == NULL)
+		return -1;
 
 	ac->type = act;
-	ac->percentile = (unsigned char)pctl;
+	ac->percentile = pctl;
 	ac->metric = strdup(metric);
+	if (ac->metric == NULL)
+		return -1;
 	memset(ac->invocations_ht, 0, sizeof(ac->invocations_ht));
 	ac->entries_needed = store;
 	pthread_rwlock_init(&ac->invlock, NULL);
@@ -234,7 +212,8 @@ aggregator_putmetric(
 			((omhash >> AGGR_HT_POW_SIZE) ^ omhash) &
 			(((unsigned int)1 << AGGR_HT_POW_SIZE) - 1);
 
-		pthread_rwlock_wrlock(&compute->invlock);
+		/* first "cheap" search for an invocation, using a read-lock */
+		pthread_rwlock_rdlock(&compute->invlock);
 		invocation = compute->invocations_ht[omhtbucket];
 		for (; invocation != NULL; invocation = invocation->next)
 			if (invocation->hash == omhash &&
@@ -245,60 +224,75 @@ aggregator_putmetric(
 			long long int i;
 			time_t now;
 
-			if ((invocation = malloc(sizeof(*invocation))) == NULL) {
-				logerr("aggregator: out of memory creating %s from %s",
-						ometric, metric);
-				pthread_rwlock_unlock(&compute->invlock);
-				continue;
-			}
-			if ((invocation->metric = strdup(ometric)) == NULL) {
-				logerr("aggregator: out of memory creating %s from %s",
-						ometric, metric);
-				free(invocation);
-				pthread_rwlock_unlock(&compute->invlock);
-				continue;
-			}
-			invocation->hash = omhash;
+			/* swap the lock for an expensive exclusive lock to
+			 * add/update the invocation */
+			pthread_rwlock_unlock(&compute->invlock);
+			pthread_rwlock_wrlock(&compute->invlock);
 
-			/* Start buckets in the past such that expiry time
-			 * conditions are met.  Add a splay to the expiry time to
-			 * avoid a thundering herd of expirations when the
-			 * aggregator is spammed with metrics, e.g. right after
-			 * startup when other relays flush their queues.  This
-			 * approach shouldn't affect the timing of the buckets as
-			 * requested in issue #72.
-			 * For consistency with other tools/old carbon-aggregator
-			 * align the buckets to interval boundaries such that it is
-			 * predictable what intervals will be taken, issue #104. */
-			time(&now);
-			now = ((now - s->expire) / s->interval) * s->interval;
-			invocation->expire = s->expire + (rand() % s->interval);
+			/* recheck if invocation is still NULL */
+			invocation = compute->invocations_ht[omhtbucket];
+			for (; invocation != NULL; invocation = invocation->next)
+				if (invocation->hash == omhash &&
+						strcmp(ometric, invocation->metric) == 0)  /* match */
+					break;
 
-			/* allocate enough buckets to hold the past + future */
-			invocation->buckets =
-				malloc(sizeof(struct _aggr_bucket) * s->bucketcnt);
-			if (invocation->buckets == NULL) {
-				logerr("aggregator: out of memory creating %s from %s",
-						ometric, metric);
-				free(invocation->metric);
-				free(invocation);
-				pthread_rwlock_unlock(&compute->invlock);
-				continue;
-			}
-			for (i = 0; i < s->bucketcnt; i++) {
-				invocation->buckets[i].start =
-					now + (i * (long long int)s->interval);
-				invocation->buckets[i].cnt = 0;
-				invocation->buckets[i].entries.size = 0;
-				invocation->buckets[i].entries.values = NULL;
-			}
+			if (invocation == NULL) {
+				if ((invocation = malloc(sizeof(*invocation))) == NULL) {
+					logerr("aggregator: out of memory creating %s from %s",
+							ometric, metric);
+					pthread_rwlock_unlock(&compute->invlock);
+					continue;
+				}
+				if ((invocation->metric = strdup(ometric)) == NULL) {
+					logerr("aggregator: out of memory creating %s from %s",
+							ometric, metric);
+					free(invocation);
+					pthread_rwlock_unlock(&compute->invlock);
+					continue;
+				}
+				invocation->hash = omhash;
 
-			invocation->next = compute->invocations_ht[omhtbucket];
-			compute->invocations_ht[omhtbucket] = invocation;
+				/* Start buckets in the past such that expiry time
+				 * conditions are met.  Add a splay to the expiry time to
+				 * avoid a thundering herd of expirations when the
+				 * aggregator is spammed with metrics, e.g. right after
+				 * startup when other relays flush their queues.  This
+				 * approach shouldn't affect the timing of the buckets as
+				 * requested in issue #72.
+				 * For consistency with other tools/old carbon-aggregator
+				 * align the buckets to interval boundaries such that it is
+				 * predictable what intervals will be taken, issue #104. */
+				time(&now);
+				now = ((now - s->expire) / s->interval) * s->interval;
+				invocation->expire = s->expire + (rand() % s->interval);
+
+				/* allocate enough buckets to hold the past + future */
+				invocation->buckets =
+					malloc(sizeof(struct _aggr_bucket) * s->bucketcnt);
+				if (invocation->buckets == NULL) {
+					logerr("aggregator: out of memory creating %s from %s",
+							ometric, metric);
+					free(invocation->metric);
+					free(invocation);
+					pthread_rwlock_unlock(&compute->invlock);
+					continue;
+				}
+				for (i = 0; i < s->bucketcnt; i++) {
+					bucket = &invocation->buckets[i];
+					pthread_mutex_init(&bucket->bcklock, NULL);
+					bucket->state = A_OPEN;
+					bucket->start = now + (i * (long long int)s->interval);
+					bucket->cnt = 0;
+					bucket->entries.size = 0;
+					bucket->entries.values = NULL;
+				}
+
+				invocation->next = compute->invocations_ht[omhtbucket];
+				compute->invocations_ht[omhtbucket] = invocation;
+			}
 		}
 
 		/* finally, try to do the maths */
-
 		itime = epoch - invocation->buckets[0].start;
 		if (itime < 0) {
 			/* drop too old metric */
@@ -319,6 +313,17 @@ aggregator_putmetric(
 		}
 
 		bucket = &invocation->buckets[itime / s->interval];
+		if (__sync_bool_compare_and_swap(&bucket->state, A_EXPIRE, A_EXPIRE)) {
+			/* drop too old metric */
+			__sync_add_and_fetch(&s->dropped, 1);
+			pthread_rwlock_unlock(&compute->invlock);
+			continue;
+		}
+
+		/* Since we have for most cases just a read-lock on invocation
+		 * here, run the bucket modification in a critical section of
+		 * its own. */
+		pthread_mutex_lock(&bucket->bcklock);
 		if (bucket->cnt == 0) {
 			bucket->sum = val;
 			bucket->max = val;
@@ -348,7 +353,7 @@ aggregator_putmetric(
 				entries->values[bucket->cnt] = val;
 		}
 		bucket->cnt++;
-
+		pthread_mutex_unlock(&bucket->bcklock);
 		pthread_rwlock_unlock(&compute->invlock);
 	}
 
@@ -361,6 +366,102 @@ cmp_entry(const void *l, const void *r)
 	return *(const double *)l - *(const double *)r;
 }
 
+static inline void
+write_metric(
+		aggregator *s,
+		struct _aggr_bucket *b,
+		struct _aggr_computes *c,
+		char *invmetric)
+{
+	char metric[METRIC_BUFSIZ];
+	double *values;
+	size_t len = 0;
+	long long int ts = 0;
+	int k;
+
+	switch (s->tswhen) {
+		case TS_START:
+			ts = b->start;
+			break;
+		case TS_MIDDLE:
+			ts = b->start + (s->interval / 2);
+			break;
+		case TS_END:
+			ts = b->start + s->interval;
+			break;
+		default:
+			assert(0);
+	}
+
+	pthread_mutex_lock(&b->bcklock);
+	switch (c->type) {
+		case SUM:
+			len = snprintf(metric, sizeof(metric),
+					"%s %f %lld\n", invmetric, b->sum, ts);
+			break;
+		case CNT:
+			len = snprintf(metric, sizeof(metric),
+					"%s %zu %lld\n", invmetric, b->cnt, ts);
+			break;
+		case MAX:
+			len = snprintf(metric, sizeof(metric),
+					"%s %f %lld\n", invmetric, b->max, ts);
+			break;
+		case MIN:
+			len = snprintf(metric, sizeof(metric),
+					"%s %f %lld\n", invmetric, b->min, ts);
+			break;
+		case AVG:
+			len = snprintf(metric, sizeof(metric),
+					"%s %f %lld\n", invmetric, b->sum / (double)b->cnt, ts);
+			break;
+		case MEDN:
+			/* median == 50th percentile */
+		case PCTL:
+			/* nearest rank method */
+			k = (int)(((double)c->percentile/100.0 * (double)b->cnt) + 0.9);
+			values = b->entries.values;
+			/* TODO: lazy approach, in case of 1 (first/last) or 2
+			 * buckets distance we could do a forward run picking the
+			 * max entries and returning that iso sorting the full array */
+			qsort(values, b->cnt, sizeof(double), cmp_entry);
+			len = snprintf(metric, sizeof(metric),
+					"%s %f %lld\n", invmetric, values[k - 1], ts);
+			break;
+		case VAR:
+		case SDEV: {
+			double avg = b->sum / (double)b->cnt;
+			double ksum = 0;
+			values = b->entries.values;
+			for (k = 0; k < b->cnt; k++)
+			   ksum += pow(values[k] - avg, 2);
+			ksum /= (double)b->cnt;
+			len = snprintf(metric, sizeof(metric), "%s %f %lld\n",
+				   invmetric, c->type == VAR ? ksum : sqrt(ksum), ts);
+		}	break;
+		default:
+			assert(0);  /* for compiler (len) */
+	}
+	pthread_mutex_unlock(&b->bcklock);
+
+	/* perform the write outside of any lock */
+	pthread_rwlock_unlock(&c->invlock);
+
+	ts = write(s->fd, metric, len);
+	if (ts < 0) {
+		logerr("aggregator: failed to write to "
+				"pipe (fd=%d): %s\n",
+				s->fd, strerror(errno));
+		__sync_add_and_fetch(&s->dropped, 1);
+	} else if (ts < len) {
+		logerr("aggregator: uncomplete write on "
+				"pipe (fd=%d)\n", s->fd);
+		__sync_add_and_fetch(&s->dropped, 1);
+	} else {
+		__sync_add_and_fetch(&s->sent, 1);
+	}
+}
+
 /**
  * Checks if the oldest bucket should be expired, if so, sends out
  * computed aggregate metrics and moves the bucket to the end of the
@@ -380,12 +481,9 @@ aggregator_expire(void *sub)
 	double *values;
 	size_t len = 0;
 	int i;
-	size_t k;
 	unsigned char j;
 	int work;
-	char metric[METRIC_BUFSIZ];
 	char isempty;
-	long long int ts = 0;
 
 	while (1) {
 		work = 0;
@@ -394,118 +492,51 @@ aggregator_expire(void *sub)
 			/* send metrics for buckets that are completely past the
 			 * expiry time, unless we are shutting down, then send
 			 * metrics for all buckets that have completed */
-			now = time(NULL) + (__sync_bool_compare_and_swap(&keep_running, 1, 1) ? 0 : s->expire - s->interval);
+			now = time(NULL) + (__sync_bool_compare_and_swap(
+						&keep_running, 1, 1) ? 0 : s->expire - s->interval);
+			for (c = s->computes; c != NULL; c = c->next) {
+				pthread_rwlock_rdlock(&c->invlock);
+				for (i = 0; i < (1 << AGGR_HT_POW_SIZE); i++) {
+					isempty = 0;
+					for (inv = c->invocations_ht[i];
+							inv != NULL;
+							inv = inv->next)
+					{
+						for (b = &inv->buckets[0];
+							b->start + (__sync_bool_compare_and_swap(
+										&keep_running, 1, 1) ?
+								 		inv->expire : s->expire) < now; b++)
+						{
+							if (!__sync_bool_compare_and_swap(
+										&b->state, A_OPEN, A_EXPIRE))
+								continue;
+
+							/* avoid emitting empty/unitialised data */
+							if (b->cnt > 0) {
+								/* write_metric unlocks before doing the
+								 * write to avoid any blocks */
+								write_metric(s, b, c, inv->metric);
+								pthread_rwlock_rdlock(&c->invlock);
+							}
+
+							work++;
+						}
+					}
+				}
+				pthread_rwlock_unlock(&c->invlock);
+			}
+
+			/* now do some cleanup for A_EXPIREd buckets in an exclusive
+			 * lock */
 			for (c = s->computes; c != NULL; c = c->next) {
 				pthread_rwlock_wrlock(&c->invlock);
 				for (i = 0; i < (1 << AGGR_HT_POW_SIZE); i++) {
 					lastinv = NULL;
 					isempty = 0;
 					for (inv = c->invocations_ht[i]; inv != NULL; ) {
-						while (inv->buckets[0].start +
-								(__sync_bool_compare_and_swap(&keep_running, 1, 1) ? inv->expire : s->expire) < now)
-						{
-							/* yay, let's produce something cool */
-							b = &inv->buckets[0];
-							/* avoid emitting empty/unitialised data */
-							isempty = b->cnt == 0;
-							if (!isempty) {
-								switch (s->tswhen) {
-									case TS_START:
-										ts = b->start;
-										break;
-									case TS_MIDDLE:
-										ts = b->start + (s->interval / 2);
-										break;
-									case TS_END:
-										ts = b->start + s->interval;
-										break;
-									default:
-										assert(0);
-								}
-								switch (c->type) {
-									case SUM:
-										len = snprintf(metric, sizeof(metric),
-												"%s %f %lld\n",
-												inv->metric, b->sum, ts);
-										break;
-									case CNT:
-										len = snprintf(metric, sizeof(metric),
-												"%s %zu %lld\n",
-												inv->metric, b->cnt, ts);
-										break;
-									case MAX:
-										len = snprintf(metric, sizeof(metric),
-												"%s %f %lld\n",
-												inv->metric, b->max, ts);
-										break;
-									case MIN:
-										len = snprintf(metric, sizeof(metric),
-												"%s %f %lld\n",
-												inv->metric, b->min, ts);
-										break;
-									case AVG:
-										len = snprintf(metric, sizeof(metric),
-												"%s %f %lld\n",
-												inv->metric,
-												b->sum / (double)b->cnt, ts);
-										break;
-									case MEDN:
-										/* median == 50th percentile */
-									case PCTL:
-										/* nearest rank method */
-										k = (int)(((double)c->percentile/100.0 *
-														(double)b->cnt) + 0.9);
-										values = b->entries.values;
-										/* TODO: lazy approach, in case
-										 * of 1 (first/last) or 2 buckets
-										 * distance we could do a
-										 * forward run picking the max
-										 * entries and returning that
-										 * iso sorting the full array */
-										qsort(values, b->cnt,
-												sizeof(double), cmp_entry);
-										len = snprintf(metric, sizeof(metric),
-												"%s %f %lld\n",
-												inv->metric,
-												values[k - 1],
-												ts);
-										break;
-									case VAR:
-									case SDEV: {
-										double avg = b->sum / (double)b->cnt;
-										double ksum = 0;
-										values = b->entries.values;
-										for (k = 0; k < b->cnt; k++)
-											ksum += pow(values[k] - avg, 2);
-										ksum /= (double)b->cnt;
-										len = snprintf(metric, sizeof(metric),
-												"%s %f %lld\n",
-												inv->metric,
-												c->type == VAR ? ksum :
-													sqrt(ksum),
-												ts);
-									}	break;
-									default:
-									   assert(0);  /* for compiler (len) */
-								}
-								ts = write(s->fd, metric, len);
-								if (ts < 0) {
-									logerr("aggregator: failed to write to "
-											"pipe (fd=%d): %s\n",
-											s->fd, strerror(errno));
-									__sync_add_and_fetch(&s->dropped, 1);
-								} else if (ts < len) {
-									logerr("aggregator: uncomplete write on "
-											"pipe (fd=%d)\n", s->fd);
-									__sync_add_and_fetch(&s->dropped, 1);
-								} else {
-									__sync_add_and_fetch(&s->sent, 1);
-								}
-							}
-
-							/* move the bucket to the end, to make room for
-							 * new ones */
-							b = &inv->buckets[0];
+						/* move expired buckets to the end */
+						b = &inv->buckets[0];
+						while (b->state == A_EXPIRE) {
 							len = b->entries.size;
 							values = b->entries.values;
 							memmove(b, &inv->buckets[1],
@@ -517,17 +548,16 @@ aggregator_expire(void *sub)
 								s->interval;
 							b->entries.size = len;
 							b->entries.values = values;
+							b->state = A_OPEN;
 
-							work++;
+							b = &inv->buckets[0];
 						}
-
-						if (isempty) {
-							/* see if the remaining buckets are empty too */
-							for (j = 0; j < s->bucketcnt; j++) {
-								if (inv->buckets[j].cnt != 0) {
-									isempty = 0;
-									break;
-								}
+						/* see if the entire invocation is empty */
+						isempty = 1;
+						for (j = 0; j < s->bucketcnt; j++) {
+							if (inv->buckets[j].cnt != 0) {
+								isempty = 0;
+								break;
 							}
 						}
 						if (isempty) {
@@ -537,6 +567,8 @@ aggregator_expire(void *sub)
 									if (inv->buckets[j].entries.values)
 										free(inv->buckets[j].entries.values);
 							free(inv->metric);
+							for (j = 0; j < s->bucketcnt; j++)
+								pthread_mutex_destroy(&inv->buckets[j].bcklock);
 							free(inv->buckets);
 							if (lastinv != NULL) {
 								lastinv->next = inv->next;
@@ -582,6 +614,8 @@ aggregator_expire(void *sub)
 						for (j = 0; j < s->bucketcnt; j++)
 							if (inv->buckets[j].entries.values)
 								free(inv->buckets[j].entries.values);
+					for (j = 0; j < s->bucketcnt; j++)
+						pthread_mutex_destroy(&inv->buckets[j].bcklock);
 					free(inv->buckets);
 
 					inv = invocation->next;
