@@ -43,6 +43,8 @@ typedef struct _aggregator {
 			unsigned int hash;  /* to speed up matching */
 			unsigned short expire;  /* expire + splay */
 			struct _aggr_bucket {
+				pthread_mutex_t bcklock;
+				enum { A_OPEN, A_EXPIRE } state;
 				long long int start;
 				size_t cnt;
 				double sum;
@@ -64,7 +66,7 @@ typedef struct _aggregator {
 } aggregator;
 
 aggregator *aggregator_new(unsigned int interval, unsigned int expire, enum _aggr_timestamp tswhen);
-char aggregator_add_compute(aggregator *s, const char *metric, const char *type);
+char aggregator_add_compute(aggregator *s, const char *metric, enum _aggr_compute_type type, unsigned char pctl);
 void aggregator_set_stub(aggregator *s, const char *stubname);
 void aggregator_putmetric(aggregator *s, const char *metric, const char *firstspace, size_t nmatch, regmatch_t *pmatch);
 int aggregator_start(aggregator *aggrs);
