@@ -32,8 +32,6 @@ static dispatcher **dispatchers;
 static char debug = 0;
 static pthread_t collectorid;
 static char keep_running = 1;
-int collector_interval = 60;
-static char cluster_refresh_pending = 0;
 static router *pending_router = NULL;
 
 /**
@@ -61,6 +59,8 @@ collector_runner(void *s)
 	time_t nextcycle;
 	char destbuf[1024];  /* sort of POSIX_MAX_PATH */
 	char *p;
+	int collector_interval = 60;
+	router *prtr;
 	size_t numaggregators = 0;
 	aggregator *aggrs = NULL;
 	server *submission = (server *)s;
@@ -68,44 +68,17 @@ collector_runner(void *s)
 	char metric[METRIC_BUFSIZ];
 	char *m = NULL;
 	size_t sizem = 0;
-	size_t (*s_ticks)(server *);
-	size_t (*s_metrics)(server *);
-	size_t (*s_stalls)(server *);
-	size_t (*s_dropped)(server *);
-	size_t (*d_ticks)(dispatcher *);
-	size_t (*d_metrics)(dispatcher *);
-	size_t (*d_blackholes)(dispatcher *);
-	size_t (*d_sleeps)(dispatcher *);
-	size_t (*a_received)(aggregator *);
-	size_t (*a_sent)(aggregator *);
-	size_t (*a_dropped)(aggregator *);
-
-	/* setup functions to target what the user wants */
-	if (debug & 2) {
-		s_ticks = server_get_ticks_sub;
-		s_metrics = server_get_metrics_sub;
-		s_stalls = server_get_stalls_sub;
-		s_dropped = server_get_dropped_sub;
-		d_ticks = dispatch_get_ticks_sub;
-		d_metrics = dispatch_get_metrics_sub;
-		d_blackholes = dispatch_get_blackholes_sub;
-		d_sleeps = dispatch_get_sleeps_sub;
-		a_received = aggregator_get_received_sub;
-		a_sent = aggregator_get_sent_sub;
-		a_dropped = aggregator_get_dropped_sub;
-	} else {
-		s_ticks = server_get_ticks;
-		s_metrics = server_get_metrics;
-		s_stalls = server_get_stalls;
-		s_dropped = server_get_dropped;
-		d_ticks = dispatch_get_ticks;
-		d_metrics = dispatch_get_metrics;
-		d_blackholes = dispatch_get_blackholes;
-		d_sleeps = dispatch_get_sleeps;
-		a_received = aggregator_get_received;
-		a_sent = aggregator_get_sent;
-		a_dropped = aggregator_get_dropped;
-	}
+	size_t (*s_ticks)(server *) = NULL;
+	size_t (*s_metrics)(server *) = NULL;
+	size_t (*s_stalls)(server *) = NULL;
+	size_t (*s_dropped)(server *) = NULL;
+	size_t (*d_ticks)(dispatcher *) = NULL;
+	size_t (*d_metrics)(dispatcher *) = NULL;
+	size_t (*d_blackholes)(dispatcher *) = NULL;
+	size_t (*d_sleeps)(dispatcher *) = NULL;
+	size_t (*a_received)(aggregator *) = NULL;
+	size_t (*a_sent)(aggregator *) = NULL;
+	size_t (*a_dropped)(aggregator *) = NULL;
 
 #define send(metric) \
 	if (debug & 1) \
@@ -115,26 +88,57 @@ collector_runner(void *s)
 
 	nextcycle = time(NULL) + collector_interval;
 	while (__sync_bool_compare_and_swap(&keep_running, 1, 1)) {
-		if (cluster_refresh_pending) {
-			char *stub = router_getcollectorstub(pending_router);
-			server **newservers = router_getservers(pending_router);
+		if ((prtr = __sync_add_and_fetch(&pending_router, 0)) != NULL) {
+			char *stub;
+			char *prefix;
+			server **newservers;
+
 			if (srvs != NULL)
 				free(srvs);
-			srvs = newservers;
-			aggrs = router_getaggregators(pending_router);
+			stub = router_getcollectorstub(prtr);
+			prefix = router_getcollectorprefix(prtr);
+			srvs = newservers = router_getservers(prtr);
+			aggrs = router_getaggregators(prtr);
 			numaggregators = aggregator_numaggregators(aggrs);
+			collector_interval = router_getcollectorinterval(prtr);
+			nextcycle = time(NULL) + collector_interval;
 
-			/* prepare hostname for graphite metrics */
-			snprintf(metric, sizeof(metric), "%scarbon.relays.%s",
-					stub == NULL ? "" : stub, relay_hostname);
-			for (m = metric + strlen("carbon.relays."); *m != '\0'; m++)
-				if (*m == '.')
-					*m = '_';
+			/* setup functions to target what the user wants */
+			if (router_getcollectormode(prtr) == SUB) {
+				s_ticks = server_get_ticks_sub;
+				s_metrics = server_get_metrics_sub;
+				s_stalls = server_get_stalls_sub;
+				s_dropped = server_get_dropped_sub;
+				d_ticks = dispatch_get_ticks_sub;
+				d_metrics = dispatch_get_metrics_sub;
+				d_blackholes = dispatch_get_blackholes_sub;
+				d_sleeps = dispatch_get_sleeps_sub;
+				a_received = aggregator_get_received_sub;
+				a_sent = aggregator_get_sent_sub;
+				a_dropped = aggregator_get_dropped_sub;
+			} else {
+				s_ticks = server_get_ticks;
+				s_metrics = server_get_metrics;
+				s_stalls = server_get_stalls;
+				s_dropped = server_get_dropped;
+				d_ticks = dispatch_get_ticks;
+				d_metrics = dispatch_get_metrics;
+				d_blackholes = dispatch_get_blackholes;
+				d_sleeps = dispatch_get_sleeps;
+				a_received = aggregator_get_received;
+				a_sent = aggregator_get_sent;
+				a_dropped = aggregator_get_dropped;
+			}
+
+			/* prepare prefix for graphite metrics */
+			i = snprintf(metric, sizeof(metric), "%s%s",
+					stub == NULL ? "" : stub, prefix);
+			m = metric + i;
 			*m++ = '.';
 			*m = '\0';
 			sizem = sizeof(metric) - (m - metric);
 
-			cluster_refresh_pending = 0;
+			(void)__sync_and_and_fetch(&pending_router, NULL);
 		}
 		assert(srvs != NULL);
 		sleep(1);
@@ -301,18 +305,21 @@ collector_writer(void *unused)
 	size_t lastconn = 0;
 	size_t lastdisc = 0;
 	size_t numaggregators = 0;
+	router *prtr;
 	server **srvs = NULL;
 	aggregator *aggrs = NULL;
+	int collector_interval = 60;
 
-	while (keep_running) {
-		if (cluster_refresh_pending) {
-			server **newservers = router_getservers(pending_router);
+	while (__sync_bool_compare_and_swap(&keep_running, 1, 1)) {
+		if ((prtr = __sync_add_and_fetch(&pending_router, 0)) != NULL) {
+			server **newservers = router_getservers(prtr);
 			if (srvs != NULL)
 				free(srvs);
 			srvs = newservers;
-			aggrs = router_getaggregators(pending_router);
+			aggrs = router_getaggregators(prtr);
 			numaggregators = aggregator_numaggregators(aggrs);
-			cluster_refresh_pending = 0;
+			collector_interval = router_getcollectorinterval(prtr);
+			(void)__sync_and_and_fetch(&pending_router, NULL);
 		}
 		assert(srvs != NULL);
 		sleep(1);
@@ -410,6 +417,10 @@ collector_writer(void *unused)
 
 		i = 0;
 	}
+
+	if (srvs != NULL)
+		free(srvs);
+
 	return NULL;
 }
 
@@ -420,8 +431,7 @@ collector_writer(void *unused)
 inline void
 collector_schedulereload(router *rtr)
 {
-	pending_router = rtr;
-	cluster_refresh_pending = 1;
+	__sync_bool_compare_and_swap(&pending_router, NULL, rtr);
 }
 
 /**
@@ -431,7 +441,7 @@ collector_schedulereload(router *rtr)
 inline char
 collector_reloadcomplete(void)
 {
-	return cluster_refresh_pending == 0;
+	return __sync_bool_compare_and_swap(&pending_router, NULL, NULL);
 }
 
 /**
