@@ -36,6 +36,16 @@
 #include "collector.h"
 #include "dispatcher.h"
 
+#ifdef HAVE_GZIP
+#include <zlib.h>
+#endif
+#ifdef HAVE_LZ4
+#include <lz4.h>
+#endif
+#ifdef HAVE_SSL
+#include <openssl/ssl.h>
+#endif
+
 enum conntype {
 	LISTENER,
 	CONNECTION
@@ -43,6 +53,9 @@ enum conntype {
 
 typedef struct _connection {
 	int sock;
+	void *strm;
+	ssize_t (*strmread)(void *, void *, size_t);  /* read func */
+	int (*strmclose)(void *);
 	char takenby;   /* -2: being setup, -1: free, 0: not taken, >0: tid */
 	char srcaddr[24];  /* string representation of source address */
 	char buf[METRIC_BUFSIZ];
@@ -59,6 +72,21 @@ typedef struct _connection {
 	char isudp:1;
 } connection;
 
+typedef struct _z_strm {
+	int sock;
+	union {
+#ifdef HAVE_GZIP
+		z_stream z;
+#endif
+#ifdef HAVE_LZ4
+		LZ4_streamDecode_t *lz;
+#endif
+		char dummy;
+	} hdl;
+	char ibuf[METRIC_BUFSIZ];
+	unsigned short ipos;
+} z_strm;
+
 struct _dispatcher {
 	pthread_t tid;
 	enum conntype type;
@@ -75,11 +103,11 @@ struct _dispatcher {
 	router *rtr;
 	router *pending_rtr;
 	char route_refresh_pending;  /* full byte for atomic access */
-	char hold:1;
+	char hold;  /* full byte for atomic access */
 	char *allowed_chars;
 };
 
-static connection **listeners = NULL;
+static listener **listeners = NULL;
 static connection *connections = NULL;
 static size_t connectionslen = 0;
 pthread_rwlock_t listenerslock = PTHREAD_RWLOCK_INITIALIZER;
@@ -88,6 +116,163 @@ static size_t acceptedconnections = 0;
 static size_t closedconnections = 0;
 static unsigned int sockbufsize = 0;
 
+/* connection specific readers and closers */
+
+/* ordinary socket */
+static inline ssize_t
+sockread(void *strm, void *buf, size_t sze)
+{
+	return read(*((int *)strm), buf, sze);
+}
+
+static inline int
+sockclose(void *strm)
+{
+	int ret = close(*((int *)strm));
+	free(strm);
+	return ret;
+}
+
+#ifdef HAVE_GZIP
+/* gzip wrapped socket */
+static inline ssize_t
+gzipread(void *strm, void *buf, size_t sze)
+{
+	z_stream *zstrm = &(((z_strm *)strm)->hdl.z);
+	char *ibuf = ((z_strm *)strm)->ibuf;
+	int ret;
+	int iret;
+	int err = 0;
+
+	/* read any available data, if it fits */
+	ret = read(((z_strm *)strm)->sock,
+			ibuf + ((z_strm *)strm)->ipos,
+			METRIC_BUFSIZ - ((z_strm *)strm)->ipos);
+	if (ret > 0) {
+		((z_strm *)strm)->ipos += ret;
+	} else if (ret < 0) {
+		err = errno;
+	}
+
+	zstrm->next_in = (Bytef *)ibuf;
+	zstrm->avail_in = (uInt)(((z_strm *)strm)->ipos);
+	zstrm->next_out = (Bytef *)buf;
+	zstrm->avail_out = (uInt)sze;
+
+	iret = inflate(zstrm, Z_SYNC_FLUSH);
+
+	/* if deflate read something, update our ibuf */
+	if ((Bytef *)ibuf != zstrm->next_in) {
+		memmove(ibuf, zstrm->next_in, zstrm->avail_in);
+		((z_strm *)strm)->ipos = zstrm->avail_in;
+	}
+
+	switch (iret) {
+		case Z_OK:  /* progress has been made */
+			/* calculate the "returned" bytes */
+			iret = sze - zstrm->avail_out;
+			if (iret == 0) {
+				/* something was done, so must have been reading input */
+				errno = EAGAIN;
+				return -1;
+			}
+			break;
+		case Z_STREAM_END:  /* everything uncompressed, nothing pending */
+			iret = sze - zstrm->avail_out;
+			break;
+		case Z_DATA_ERROR:  /* corrupt input */
+			inflateSync(zstrm);
+			/* return isn't much of interest, we will call inflate next
+			 * time and sync again if it still fails */
+			errno = err ? err : EAGAIN;
+			return -1;
+			break;
+		case Z_MEM_ERROR:  /* out of memory */
+			logerr("out of memory during read of gzip stream\n");
+			errno = ENOMEM;
+			return -1;
+		case Z_BUF_ERROR:  /* output buffer full or nothing to read */
+			/* if this happens, zlib still can't write even though last
+			 * call should have increased the buffer size by consuming
+			 * whatever was valid, so if we're stuck here, the situation
+			 * isn't going to improve, and we better bail out */
+			errno = err ? err : EMSGSIZE;
+			return -1;
+	}
+
+	if (iret == 0 && ret != 0) {
+		errno = EAGAIN;
+		iret = -1;
+	}
+
+	return (ssize_t)iret;
+}
+
+static inline int
+gzipclose(void *strm)
+{
+	int ret = close(((z_strm *)strm)->sock);
+	inflateEnd(&(((z_strm *)strm)->hdl.z));
+	free(strm);
+	return ret;
+}
+#endif
+
+#ifdef HAVE_LZ4
+/* lz4 wrapped socket */
+static inline ssize_t
+lzread(void *strm, void *buf, size_t sze)
+{
+	char *ibuf = ((z_strm *)strm)->ibuf;
+	int ret;
+
+	/* read any available data, if it fits */
+	ret = read(((z_strm *)strm)->sock,
+			ibuf + ((z_strm *)strm)->ipos,
+			METRIC_BUFSIZ - ((z_strm *)strm)->ipos);
+	if (ret > 0) {
+		((z_strm *)strm)->ipos += ret;
+	} else if (ret < 0) {
+		return -1;
+	}
+
+	ret = LZ4_decompress_safe_continue(((z_strm *)strm)->hdl.lz,
+			ibuf, buf, ((z_strm *)strm)->ipos, sze);
+
+	/* if we decompressed something, update our ibuf */
+	if (ret > 0) {
+		((z_strm *)strm)->ipos = ((z_strm *)strm)->ipos - ret;
+		memmove(ibuf, ibuf + ret, ((z_strm *)strm)->ipos);
+	}
+
+	return (ssize_t)ret;
+}
+
+static inline int
+lzclose(void *strm)
+{
+	int ret = LZ4_freeStreamDecode(((z_strm *)strm)->hdl.lz);
+	free(strm);
+	return ret;
+}
+#endif
+
+#ifdef HAVE_SSL
+/* (Open|Libre)SSL wrapped socket */
+static inline ssize_t
+sslread(void *strm, void *buf, size_t sze)
+{
+	return (ssize_t)SSL_read((SSL *)strm, buf, (int)sze);
+}
+
+static inline int
+sslclose(void *strm)
+{
+	int sock = SSL_get_fd((SSL *)strm);
+	SSL_free((SSL *)strm);
+	return close(sock);
+}
+#endif
 
 /**
  * Helper function to try and be helpful to the user.  If errno
@@ -119,34 +304,36 @@ dispatch_check_rlimit_and_warn(void)
  * Listener sockets are those which need to be accept()-ed on.
  */
 int
-dispatch_addlistener(int sock)
+dispatch_addlistener(listener *lsnr)
 {
-	connection *newconn;
 	int c;
+	int *socks;
+
+	if (lsnr->ctype == CON_UDP) {
+		/* Adds a pseudo-listener for datagram (UDP) sockets, which is
+		 * pseudo, for in fact it adds a new connection, but makes sure
+		 * that connection won't be closed after being idle, and won't
+		 * count that connection as an incoming connection either. */
+		for (socks = lsnr->socks; *socks != -1; socks++) {
+			c = dispatch_addconnection(*socks, lsnr);
+
+			if (c == -1)
+				return 1;
+
+			connections[c].noexpire = 1;
+			connections[c].isudp = 1;
+			acceptedconnections--;
+		}
+
+		return 0;
+	}
 
 	pthread_rwlock_wrlock(&listenerslock);
-	if (listeners == NULL) {
-		if ((listeners = malloc(sizeof(connection *) * MAX_LISTENERS)) == NULL)
-		{
-			pthread_rwlock_unlock(&listenerslock);
-			return 1;
-		}
-		for (c = 0; c < MAX_LISTENERS; c++)
-			listeners[c] = NULL;
-	}
 	for (c = 0; c < MAX_LISTENERS; c++) {
 		if (listeners[c] == NULL) {
-			newconn = malloc(sizeof(connection));
-			if (newconn == NULL) {
-				pthread_rwlock_unlock(&listenerslock);
-				return 1;
-			}
-			(void) fcntl(sock, F_SETFL, O_NONBLOCK);
-			newconn->sock = sock;
-			newconn->takenby = 0;
-			newconn->buflen = 0;
-
-			listeners[c] = newconn;
+			listeners[c] = lsnr;
+			for (socks = lsnr->socks; *socks != -1; socks++)
+				(void) fcntl(*socks, F_SETFL, O_NONBLOCK);
 			break;
 		}
 	}
@@ -167,28 +354,71 @@ dispatch_addlistener(int sock)
  * global lock.  Frequent usage of this function is not anticipated.
  */
 void
-dispatch_removelistener(int sock)
+dispatch_removelistener(listener *lsnr)
 {
 	int c;
-	connection *conn;
+	int *socks;
 
-	pthread_rwlock_wrlock(&listenerslock);
-	/* find connection */
-	for (c = 0; c < MAX_LISTENERS; c++)
-		if (listeners[c] != NULL && listeners[c]->sock == sock)
-			break;
-	if (c == MAX_LISTENERS) {
-		/* not found?!? */
-		logerr("dispatch: cannot find listener!\n");
+	if (lsnr->ctype != CON_UDP) {
+		pthread_rwlock_wrlock(&listenerslock);
+		/* find connection */
+		for (c = 0; c < MAX_LISTENERS; c++)
+			if (listeners[c] != NULL && listeners[c] == lsnr)
+				break;
+		if (c == MAX_LISTENERS) {
+			/* not found?!? */
+			logerr("dispatch: cannot find listener to remove!\n");
+			pthread_rwlock_unlock(&listenerslock);
+			return;
+		}
+		listeners[c] = NULL;
 		pthread_rwlock_unlock(&listenerslock);
-		return;
 	}
 	/* cleanup */
-	conn = listeners[c];
-	close(conn->sock);
-	free(conn);
-	listeners[c] = NULL;
-	pthread_rwlock_unlock(&listenerslock);
+#ifdef HAVE_SSL
+	if (lsnr->transport == W_SSL)
+		SSL_CTX_free(lsnr->ctx);
+#endif
+	/* acquire a write lock on connections, which is a bit wrong, but it
+	 * ensures all dispatchers are stopped while we close the sockets,
+	 * which avoids a race on the reading thereof if this is a UDP
+	 * connection */
+	pthread_rwlock_wrlock(&connectionslock);
+	for (socks = lsnr->socks; *socks != -1; socks++) {
+		close(*socks);
+		*socks = -1;
+	}
+	pthread_rwlock_unlock(&connectionslock);
+	if (lsnr->saddrs)
+		freeaddrinfo(lsnr->saddrs);
+}
+
+/**
+ * Copy over all state related things from olsnr to nlsnr and ensure
+ * olsnr can be discarded (that is, thrown away without calling
+ * dispatch_removelistener).
+ */
+void
+dispatch_transplantlistener(listener *olsnr, listener *nlsnr, router *r)
+{
+	int c;
+
+	pthread_rwlock_wrlock(&listenerslock);
+	for (c = 0; c < MAX_LISTENERS; c++) {
+		if (listeners[c] == olsnr) {
+			router_transplant_listener_socks(r, olsnr, nlsnr);
+#ifdef HAVE_SSL
+			if (nlsnr->transport == W_SSL)
+				nlsnr->ctx = olsnr->ctx;
+#endif
+			if (olsnr->saddrs) {
+				freeaddrinfo(olsnr->saddrs);
+				olsnr->saddrs = NULL;
+			}
+			listeners[c] = nlsnr;
+			break;  /* found and done */
+		}
+	}
 }
 
 #define CONNGROWSZ  1024
@@ -199,7 +429,7 @@ dispatch_removelistener(int sock)
  * Returns the connection id, or -1 if a failure occurred.
  */
 int
-dispatch_addconnection(int sock)
+dispatch_addconnection(int sock, listener *lsnr)
 {
 	size_t c;
 	struct sockaddr_in6 saddr;
@@ -218,7 +448,7 @@ dispatch_addconnection(int sock)
 		if (connectionslen > c) {
 			/* another dispatcher just extended the list */
 			pthread_rwlock_unlock(&connectionslock);
-			return dispatch_addconnection(sock);
+			return dispatch_addconnection(sock, lsnr);
 		}
 		newlst = realloc(connections,
 				sizeof(connection) * (connectionslen + CONNGROWSZ));
@@ -231,10 +461,10 @@ dispatch_addconnection(int sock)
 			return -1;
 		}
 
-		memset(&newlst[connectionslen], '\0',
-				sizeof(connection) * CONNGROWSZ);
-		for (c = connectionslen; c < connectionslen + CONNGROWSZ; c++)
+		for (c = connectionslen; c < connectionslen + CONNGROWSZ; c++) {
+			memset(&newlst[c], '\0', sizeof(connection));
 			newlst[c].takenby = -1;  /* free */
+		}
 		connections = newlst;
 		c = connectionslen;  /* for the setup code below */
 		newlst[c].takenby = -2;
@@ -267,6 +497,65 @@ dispatch_addconnection(int sock)
 			;
 	}
 	connections[c].sock = sock;
+	if (lsnr == NULL || lsnr->transport == W_PLAIN) {
+		int *strm = malloc(sizeof(connections[c].sock));
+		if (strm == NULL) {
+			logerr("cannot add new connection: "
+					"out of memory allocating stream\n");
+			__sync_bool_compare_and_swap(&(connections[c].takenby), -2, -1);
+			return -1;
+		}
+		*strm = sock;
+		connections[c].strm = strm;
+		connections[c].strmread = &sockread;
+		connections[c].strmclose = &sockclose;
+	}
+#ifdef HAVE_GZIP
+	else if (lsnr->transport == W_GZIP) {
+		z_strm *zstrm = malloc(sizeof(z_strm));
+		if (zstrm == NULL) {
+			logerr("cannot add new connection: "
+					"out of memory allocating gzip stream\n");
+			__sync_bool_compare_and_swap(&(connections[c].takenby), -2, -1);
+			return -1;
+		}
+		zstrm->hdl.z.zalloc = Z_NULL;
+		zstrm->hdl.z.zfree = Z_NULL;
+		zstrm->hdl.z.opaque = Z_NULL;
+		zstrm->ipos = 0;
+		zstrm->sock = sock;
+		inflateInit2(&(zstrm->hdl.z), 15 + 16);
+		connections[c].strm = zstrm;
+		connections[c].strmread = &gzipread;
+		connections[c].strmclose = &gzipclose;
+	}
+#endif
+#ifdef HAVE_LZ4
+	else if (lsnr->transport == W_LZ4) {
+		z_strm *lzstrm = malloc(sizeof(z_strm));
+		if (lzstrm == NULL) {
+			logerr("cannot add new connection: "
+					"out of memory allocating lz4 stream\n");
+			__sync_bool_compare_and_swap(&(connections[c].takenby), -2, -1);
+			return -1;
+		}
+		lzstrm->sock = sock;
+		lzstrm->hdl.lz = LZ4_createStreamDecode();
+		connections[c].strm = lzstrm;
+		connections[c].strmread = &lzread;
+		connections[c].strmclose = &lzclose;
+	}
+#endif
+#ifdef HAVE_SSL
+	else if (lsnr->transport == W_SSL) {
+		connections[c].strm = SSL_new(lsnr->ctx);
+		SSL_set_fd(connections[c].strm, sock);
+		SSL_set_accept_state(connections[c].strm);
+
+		connections[c].strmread = &sslread;
+		connections[c].strmclose = &sslclose;
+	}
+#endif
 	connections[c].buflen = 0;
 	connections[c].needmore = 0;
 	connections[c].noexpire = 0;
@@ -291,7 +580,7 @@ dispatch_addconnection(int sock)
 int
 dispatch_addconnection_aggr(int sock)
 {
-	int conn = dispatch_addconnection(sock);
+	int conn = dispatch_addconnection(sock, NULL);
 
 	if (conn == -1)
 		return 1;
@@ -303,28 +592,6 @@ dispatch_addconnection_aggr(int sock)
 	return 0;
 }
 
-/**
- * Adds a pseudo-listener for datagram (UDP) sockets, which is pseudo,
- * for in fact it adds a new connection, but makes sure that connection
- * won't be closed after being idle, and won't count that connection as
- * an incoming connection either.
- */
-int
-dispatch_addlistener_udp(int sock)
-{
-	int conn = dispatch_addconnection(sock);
-
-	if (conn == -1)
-		return 1;
-
-	connections[conn].noexpire = 1;
-	connections[conn].isudp = 1;
-	acceptedconnections--;
-
-	return 0;
-}
-
-
 inline static char
 dispatch_process_dests(connection *conn, dispatcher *self, struct timeval now)
 {
@@ -334,8 +601,9 @@ dispatch_process_dests(connection *conn, dispatcher *self, struct timeval now)
 	if (conn->destlen > 0) {
 		if (conn->maxsenddelay == 0)
 			conn->maxsenddelay = ((rand() % 750) + 250) * 1000;
-		/* force after timeout */
-		force = timediff(conn->lastwork, now) > conn->maxsenddelay;
+		/* force when aggr (don't stall it) or after timeout */
+		force = conn->isaggr ? 1 :
+			timediff(conn->lastwork, now) > conn->maxsenddelay;
 		for (i = 0; i < conn->destlen; i++) {
 			tracef("dispatcher %d, connfd %d, metric %s, queueing to %s:%d\n",
 					self->id, conn->sock, conn->dests[i].metric,
@@ -344,7 +612,7 @@ dispatch_process_dests(connection *conn, dispatcher *self, struct timeval now)
 			if (server_send(conn->dests[i].dest, conn->dests[i].metric, force) == 0)
 				break;
 		}
-		if (i != conn->destlen) {
+		if (i < conn->destlen) {
 			conn->destlen -= i;
 			memmove(&conn->dests[0], &conn->dests[i],
 					(sizeof(destination) * conn->destlen));
@@ -406,7 +674,7 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 	 * left in the buffer, try to process the buffer */
 	if (
 			(!conn->needmore && conn->buflen > 0) ||
-			(len = read(conn->sock,
+			(len = conn->strmread(conn->strm,
 						conn->buf + conn->buflen,
 						(sizeof(conn->buf) - 1) - conn->buflen)) > 0
 	   )
@@ -415,6 +683,9 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 			conn->buflen += len;
 			tracef("dispatcher %d, connfd %d, read %d bytes from socket\n",
 					self->id, conn->sock, len);
+#ifdef ENABLE_TRACE
+			conn->buf[conn->buflen] = '\0';
+#endif
 		}
 
 		/* metrics look like this: metric_path value timestamp\n
@@ -438,17 +709,21 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 				}
 
 				__sync_add_and_fetch(&(self->metrics), 1);
+				/* add newline and terminate the string, we can do this
+				 * because we substract one from buf and we always store
+				 * a full metric in buf before we copy it to metric
+				 * (which is of the same size) */
 				*q++ = '\n';
-				*q = '\0';  /* can do this because we substract one from buf */
+				*q = '\0';
 
 				/* perform routing of this metric */
 				tracef("dispatcher %d, connfd %d, metric %s",
 						self->id, conn->sock, conn->metric);
 				__sync_add_and_fetch(&(self->blackholes),
 						router_route(self->rtr,
-						conn->dests, &conn->destlen, CONN_DESTS_SIZE,
-						conn->srcaddr,
-						conn->metric, firstspace));
+							conn->dests, &conn->destlen, CONN_DESTS_SIZE,
+							conn->srcaddr,
+							conn->metric, firstspace, self->id - 1));
 				tracef("dispatcher %d, connfd %d, destinations %zd\n",
 						self->id, conn->sock, conn->destlen);
 
@@ -484,11 +759,15 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 					if (*(q - 1) != *p && (q - 1) != firstspace)
 						*q++ = *p;
 				}
-			} else if (firstspace != NULL ||
-					(*p >= 'a' && *p <= 'z') ||
-					(*p >= 'A' && *p <= 'Z') ||
-					(*p >= '0' && *p <= '9') ||
-					strchr(self->allowed_chars, *p))
+			} else if (
+					*p != '\0' && (
+						firstspace != NULL ||
+						(*p >= 'a' && *p <= 'z') ||
+						(*p >= 'A' && *p <= 'Z') ||
+						(*p >= '0' && *p <= '9') ||
+						strchr(self->allowed_chars, *p)
+					)
+				)
 			{
 				/* copy char */
 				*q++ = *p;
@@ -501,7 +780,18 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 		if (lastnl != NULL) {
 			/* move remaining stuff to the front */
 			conn->buflen -= lastnl + 1 - conn->buf;
-			memmove(conn->buf, lastnl + 1, conn->buflen);
+			tracef("dispatcher %d, conn->buf: %p, lastnl: %p, diff: %zd, "
+					"conn->buflen: %d, sizeof(conn->buf): %lu, "
+					"memmove(%d, %lu, %d)\n",
+					self->id,
+					conn->buf, lastnl, lastnl - conn->buf,
+					conn->buflen, sizeof(conn->buf),
+					0, lastnl + 1 - conn->buf, conn->buflen + 1);
+			tracef("dispatcher %d, pre conn->buf: %s\n", self->id, conn->buf);
+			/* copy last NULL-byte for debug tracing */
+			memmove(conn->buf, lastnl + 1, conn->buflen + 1);
+			tracef("dispatcher %d, post conn->buf: %s\n", self->id, conn->buf);
+			lastnl = NULL;
 		}
 	}
 	if (len == -1 && (errno == EINTR ||
@@ -534,11 +824,15 @@ dispatch_connection(connection *conn, dispatcher *self, struct timeval start)
 			__sync_bool_compare_and_swap(&(conn->takenby), self->id, 0);
 
 			return len > 0;
-		} else {
+		} else if (conn->destlen == 0) {
+			tracef("dispatcher: %d, connfd: %d, len: %d [%s], disconnecting\n",
+					self->id, conn->sock, len,
+					len < 0 ? strerror(errno) : "");
 			__sync_add_and_fetch(&closedconnections, 1);
-			close(conn->sock);
+			conn->strmclose(conn->strm);
 
-			/* flag this connection as no longer in use */
+			/* flag this connection as no longer in use, unless there is
+			 * pending metrics to send */
 			__sync_bool_compare_and_swap(&(conn->takenby), self->id, -1);
 
 			return 0;
@@ -565,25 +859,29 @@ dispatch_runner(void *arg)
 	if (self->type == LISTENER) {
 		struct pollfd ufds[MAX_LISTENERS];
 		int fds;
+		int *sock;
 		while (__sync_bool_compare_and_swap(&(self->keep_running), 1, 1)) {
 			pthread_rwlock_rdlock(&listenerslock);
 			fds = 0;
 			for (c = 0; c < MAX_LISTENERS; c++) {
 				if (listeners[c] == NULL)
 					continue;
-				ufds[fds].fd = listeners[c]->sock;
-				ufds[fds].events = POLLIN;
-				fds++;
+				for (sock = listeners[c]->socks; *sock != -1; sock++) {
+					ufds[fds].fd = *sock;
+					ufds[fds].events = POLLIN;
+					fds++;
+				}
 			}
 			pthread_rwlock_unlock(&listenerslock);
 			if (poll(ufds, fds, 1000) > 0) {
-				for (c = fds - 1; c >= 0; c--) {
-					if (ufds[c].revents & POLLIN) {
+				int f;
+				for (f = fds - 1; f >= 0; f--) {
+					if (ufds[f].revents & POLLIN) {
 						int client;
 						struct sockaddr addr;
 						socklen_t addrlen = sizeof(addr);
 
-						if ((client = accept(ufds[c].fd, &addr, &addrlen)) < 0)
+						if ((client = accept(ufds[f].fd, &addr, &addrlen)) < 0)
 						{
 							logerr("dispatch: failed to "
 									"accept() new connection: %s\n",
@@ -591,7 +889,25 @@ dispatch_runner(void *arg)
 							dispatch_check_rlimit_and_warn();
 							continue;
 						}
-						if (dispatch_addconnection(client) == -1) {
+						pthread_rwlock_rdlock(&listenerslock);
+						for (c = 0; c < MAX_LISTENERS; c++) {
+							if (listeners[c] == NULL)
+								continue;
+							for (sock = listeners[c]->socks; *sock != -1; sock++) {
+								if (ufds[f].fd == *sock)
+									break;
+							}
+							if (*sock != -1)
+								break;
+						}
+						pthread_rwlock_unlock(&listenerslock);
+						if (c == MAX_LISTENERS) {
+							logerr("dispatch: could not find listener for "
+									"socket, rejecting connection\n");
+							close(client);
+							continue;
+						}
+						if (dispatch_addconnection(client, listeners[c]) == -1) {
 							close(client);
 							continue;
 						}
@@ -610,7 +926,7 @@ dispatch_runner(void *arg)
 				self->rtr = self->pending_rtr;
 				self->pending_rtr = NULL;
 				__sync_bool_compare_and_swap(&(self->route_refresh_pending), 1, 0);
-				self->hold = 0;
+				__sync_and_and_fetch(&(self->hold), 0);
 			}
 
 			gettimeofday(&start, NULL);
@@ -621,7 +937,9 @@ dispatch_runner(void *arg)
 				if (!__sync_bool_compare_and_swap(
 							&(conn->takenby), 0, self->id))
 					continue;
-				if (self->hold && !conn->isaggr) {
+				if (__sync_bool_compare_and_swap(
+							&(self->hold), 1, 1) && !conn->isaggr)
+				{
 					__sync_bool_compare_and_swap(
 							&(conn->takenby), self->id, 0);
 					continue;
@@ -654,14 +972,23 @@ dispatch_runner(void *arg)
  * Returns its handle.
  */
 static dispatcher *
-dispatch_new(char id, enum conntype type, router *r, char *allowed_chars)
+dispatch_new(
+		unsigned char id,
+		enum conntype type,
+		router *r,
+		char *allowed_chars
+	)
 {
 	dispatcher *ret = malloc(sizeof(dispatcher));
 
 	if (ret == NULL)
 		return NULL;
+	if (type == CONNECTION && r == NULL) {
+		free(ret);
+		return NULL;
+	}
 
-	ret->id = id;
+	ret->id = id + 1;  /* ensure > 0 */
 	ret->type = type;
 	ret->keep_running = 1;
 	ret->rtr = r;
@@ -692,16 +1019,31 @@ dispatch_set_bufsize(unsigned int nsockbufsize)
 	sockbufsize = nsockbufsize;
 }
 
-static char globalid = 0;
+/**
+ * Initialise the listeners array.  This is a one-time allocation that
+ * currently never is extended.  This code does no locking, as it
+ * assumes to be run before any access to listeners occur.
+ */
+char
+dispatch_init_listeners()
+{
+	int i;
+	/* once all-or-nothing allocation */
+	if ((listeners = malloc(sizeof(listener *) * MAX_LISTENERS)) == NULL)
+		return 1;
+	for (i = 0; i < MAX_LISTENERS; i++)
+		listeners[i] = NULL;
+
+	return 0;
+}
 
 /**
  * Starts a new dispatcher specialised in handling incoming connections
  * (and putting them on the queue for handling the connections).
  */
 dispatcher *
-dispatch_new_listener(void)
+dispatch_new_listener(unsigned char id)
 {
-	char id = globalid++;
 	return dispatch_new(id, LISTENER, NULL, NULL);
 }
 
@@ -710,9 +1052,8 @@ dispatch_new_listener(void)
  * existing connections.
  */
 dispatcher *
-dispatch_new_connection(router *r, char *allowed_chars)
+dispatch_new_connection(unsigned char id, router *r, char *allowed_chars)
 {
-	char id = globalid++;
 	return dispatch_new(id, CONNECTION, r, allowed_chars);
 }
 
@@ -755,7 +1096,7 @@ dispatch_free(dispatcher *d)
 inline void
 dispatch_hold(dispatcher *d)
 {
-	d->hold = 1;
+	__sync_bool_compare_and_swap(&(d->hold), 0, 1);
 }
 
 /**
