@@ -25,14 +25,15 @@
 typedef struct _dispatcher dispatcher;
 
 void dispatch_check_rlimit_and_warn(void);
-int dispatch_addlistener(int sock);
-int dispatch_addlistener_udp(int sock);
-void dispatch_removelistener(int sock);
-int dispatch_addconnection(int sock);
+int dispatch_addlistener(listener *lsnr);
+void dispatch_removelistener(listener *lsnr);
+void dispatch_transplantlistener(listener *olsnr, listener *nlsnr, router *r);
+int dispatch_addconnection(int sock, listener *lsnr);
 int dispatch_addconnection_aggr(int sock);
 void dispatch_set_bufsize(unsigned int sockbufsize);
-dispatcher *dispatch_new_listener(void);
-dispatcher *dispatch_new_connection(router *r, char *allowed_chars);
+char dispatch_init_listeners(void);
+dispatcher *dispatch_new_listener(unsigned char id);
+dispatcher *dispatch_new_connection(unsigned char id, router *r, char *allowed_chars);
 void dispatch_stop(dispatcher *d);
 void dispatch_shutdown(dispatcher *d);
 void dispatch_free(dispatcher *d);
