@@ -18,8 +18,9 @@
 #ifndef RECEPTOR_H
 #define RECEPTOR_H 1
 
-int bindlisten(int ret_stream[], int *retlen_stream, int ret_dgram[], int *retlen_dgram, const char *interface, unsigned short port, unsigned int backlog);
+#include "router.h"
 
-void destroy_usock(unsigned short port);
+int bindlisten(listener *lsnr, unsigned int backlog);
+void shutdownclose(listener *lsnr);
 
 #endif
