@@ -19,9 +19,12 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <unistd.h>
+#include <netdb.h>
 #include <string.h>
+#include <ctype.h>
 #include <signal.h>
 #include <time.h>
+#include <sys/socket.h>
 #include <errno.h>
 #include <assert.h>
 
@@ -37,6 +40,7 @@
 #include "dispatcher.h"
 #include "aggregator.h"
 #include "collector.h"
+#include "conffile.h"
 
 unsigned char keep_running = 1;
 int pending_signal = -1;
@@ -48,7 +52,12 @@ static int batchsize = 2500;
 static int queuesize = 25000;
 static int maxstalls = 4;
 static unsigned short iotimeout = 600;
+static unsigned short listenport = GRAPHITE_PORT;
+static int optimiserthreshold = 50;
 static int sockbufsize = 0;
+static int collector_interval = 60;
+static unsigned int listenbacklog = 32;
+static col_mode smode = CUM;
 static dispatcher **workers = NULL;
 static char workercnt = 0;
 static router *rtr = NULL;
@@ -92,7 +101,9 @@ relaylog(enum logdst dest, const char *fmt, ...)
 				console = 1;
 			break;
 	}
-	assert(dst != NULL);
+	/* NULL == /dev/null */
+	if (dst == NULL)
+		return 0;
 
 	time(&now);
 	localtime_r(&now, &tm_now);
@@ -130,6 +141,7 @@ do_reload(void)
 	int id;
 	FILE *newfd;
 	size_t numaggregators;
+	listener *lsnrs;
 
 	if (relay_stderr != stderr) {
 		/* try to re-open the file first, so we can still try and say
@@ -151,14 +163,14 @@ do_reload(void)
 	}
 
 	logout("reloading config from '%s'\n", config);
-	if ((newrtr = router_readconfig(NULL, config,
+	if ((newrtr = router_readconfig(NULL, config, workercnt,
 					queuesize, batchsize, maxstalls,
-					iotimeout, sockbufsize)) == NULL)
+					iotimeout, sockbufsize, listenport)) == NULL)
 	{
 		logerr("failed to read configuration '%s', aborting reload\n", config);
 		return;
 	}
-	router_optimise(newrtr);
+	router_optimise(newrtr, optimiserthreshold);
 
 	/* compare the configs first, if there is no diff, then
 	 * just refrain from reloading anything */
@@ -170,6 +182,15 @@ do_reload(void)
 		return;
 	}
 
+	/* close connections that are not in the new config first */
+	lsnrs = router_get_listeners(rtr);
+	for ( ; lsnrs != NULL; lsnrs = lsnrs->next) {
+		if (router_contains_listener(newrtr, lsnrs) == NULL) {
+			dispatch_removelistener(lsnrs);
+			shutdownclose(lsnrs);
+		}
+	}
+
 	logout("reloading collector\n");
 	collector_schedulereload(newrtr);
 	while (!collector_reloadcomplete())
@@ -205,6 +226,27 @@ do_reload(void)
 		}
 	}
 
+	/* Transplant listeners so their open state won't get lost.  Open up
+	 * connections for new listeners. */
+	lsnrs = router_get_listeners(newrtr);
+	for ( ; lsnrs != NULL; lsnrs = lsnrs->next) {
+		listener *olsnr;
+		if ((olsnr = router_contains_listener(rtr, lsnrs)) != NULL) {
+			/* ensure we copy over the opened sockets/state */
+			dispatch_transplantlistener(olsnr, lsnrs, newrtr);
+		} else {
+			if (bindlisten(lsnrs, listenbacklog) != 0) {
+				logerr("failed to setup listener, "
+						"this will impact the behaviour of the relay\n");
+				continue;
+			}
+			if (dispatch_addlistener(lsnrs) != 0) {
+				logerr("failed to listen to socket, the "
+						"listener will not be effective\n");
+			}
+		}
+	}
+
 	/* Transplant the queues for the same servers, so we keep their
 	 * queues (potentially with data) around.  To do so, we need to make
 	 * sure the servers aren't being operated on. */
@@ -305,8 +347,8 @@ do_usage(char *name, int exitcode)
 	printf("Options:\n");
 	printf("  -v  print version and exit\n");
 	printf("  -f  read <config> for clusters and routes\n");
-	printf("  -p  listen on <port> for connections, defaults to 2003\n");
-	printf("  -i  listen on <interface> for connections, defaults to all\n");
+	printf("  -p  listen on <port> for connections, defaults to %u\n",
+			GRAPHITE_PORT);
 	printf("  -l  write output to <file>, defaults to stdout/stderr\n");
 	printf("  -w  use <workers> worker threads, defaults to %d\n", get_cores());
 	printf("  -b  server send batch size, defaults to %d\n", batchsize);
@@ -314,17 +356,22 @@ do_usage(char *name, int exitcode)
 	printf("  -L  server max stalls, defaults to %d\n", maxstalls);
 	printf("  -S  statistics sending interval in seconds, defaults to 60\n");
 	printf("  -B  connection listen backlog, defaults to 32\n");
+	printf("  -U  socket receive buffer size, max/min/default values depend on OS\n");
 	printf("  -T  IO timeout in milliseconds for server connections, defaults to %d\n", iotimeout);
 	printf("  -m  send statistics like carbon-cache.py, e.g. not cumulative\n");
 	printf("  -c  characters to allow next to [A-Za-z0-9], defaults to -_:#\n");
 	printf("  -d  debug mode: currently writes statistics to log, prints hash\n"
 	       "      ring contents and matching position in test mode (-t)\n");
+#ifdef ENABLE_TRACE
+	printf("      add another -d to enable (very noisy) trace mode\n");
+#endif
 	printf("  -s  submission mode: don't add any metrics to the stream like\n"
 	       "      statistics, report drop counts and queue pressure to log\n");
 	printf("  -t  config test mode: prints rule matches from input on stdin\n");
 	printf("  -H  hostname: override hostname (used in statistics)\n");
 	printf("  -D  daemonise: run in a background\n");
 	printf("  -P  pidfile: write a pid to a specified pidfile\n");
+	printf("  -O  minimum number of rules before optimising the ruleset, default: %d\n", optimiserthreshold); 
 
 	exit(exitcode);
 }
@@ -332,26 +379,21 @@ do_usage(char *name, int exitcode)
 int
 main(int argc, char * const argv[])
 {
-	int stream_sock[] = {0, 0, 0};  /* tcp4, tcp6, UNIX */
-	int stream_socklen = sizeof(stream_sock) / sizeof(stream_sock[0]);
-	int dgram_sock[] = {0, 0};  /* udp4, udp6 */
-	int dgram_socklen = sizeof(dgram_sock) / sizeof(dgram_sock[0]);
 	char id;
-	unsigned short listenport = 2003;
-	unsigned int listenbacklog = 32;
 	int ch;
 	size_t numaggregators;
-	char *listeninterface = NULL;
 	char *allowed_chars = NULL;
-	enum { SUB, CUM } smode = CUM;
 	char *pidfile = NULL;
 	FILE *pidfile_handle = NULL;
 	aggregator *aggrs = NULL;
+	int fds[2];
+	unsigned char startup_success = 0;
+	listener *lsnrs;
 
 	if (gethostname(relay_hostname, sizeof(relay_hostname)) < 0)
 		snprintf(relay_hostname, sizeof(relay_hostname), "127.0.0.1");
 
-	while ((ch = getopt(argc, argv, ":hvdmstf:i:l:p:w:b:q:L:S:T:c:H:B:U:DP:")) != -1) {
+	while ((ch = getopt(argc, argv, ":hvdstf:l:p:w:b:q:L:T:c:H:B:U:DP:O:")) != -1) {
 		switch (ch) {
 			case 'v':
 				do_version();
@@ -366,21 +408,20 @@ main(int argc, char * const argv[])
 					mode |= MODE_DEBUG;
 				}
 				break;
-			case 'm':
-				smode = SUB;
-				break;
 			case 's':
 				mode |= MODE_SUBMISSION;
 				break;
 			case 't':
-				mode |= MODE_TEST;
+				/* -t: test interactively, -tt: test config and exit */
+				if (mode & MODE_TEST) {
+					mode |= MODE_CONFIGTEST;
+				} else {
+					mode |= MODE_TEST;
+				}
 				break;
 			case 'f':
 				config = optarg;
 				break;
-			case 'i':
-				listeninterface = optarg;
-				break;
 			case 'l':
 				relay_logfile = optarg;
 				break;
@@ -420,14 +461,6 @@ main(int argc, char * const argv[])
 					do_usage(argv[0], 1);
 				}
 				break;
-			case 'S':
-				collector_interval = atoi(optarg);
-				if (collector_interval <= 0) {
-					fprintf(stderr, "error: sending interval needs to be "
-							"a number >0\n");
-					do_usage(argv[0], 1);
-				}
-				break;
 			case 'T': {
 				int val = atoi(optarg);
 				if (val <= 0) {
@@ -483,11 +516,11 @@ main(int argc, char * const argv[])
 				}
 				if (len != sizeof(sockbufsize)) {
 					fprintf(stderr, "getsockopt returned unexpected size: %u, "
-							"expected %lu\n", len, sizeof(sockbufsize));
+							"expected %zu\n", len, sizeof(sockbufsize));
 					exit(1);
 				}
-				if (sockbufsize != (unsigned int)val)
-					fprintf(stdout, "warning: OS rejected socket bufsize\n");
+				if (sockbufsize > (unsigned int)val)
+					fprintf(stderr, "warning: OS rejected socket bufsize\n");
 				if (sock != -1)
 					close(sock);
 			}	break;
@@ -497,6 +530,10 @@ main(int argc, char * const argv[])
 			case 'P':
 				pidfile = optarg;
 				break;
+			case 'O': {
+				int val = atoi(optarg);
+				optimiserthreshold = val < 0 ? -1 : val;
+			}	break;
 			case '?':
 			case ':':
 				do_usage(argv[0], 1);
@@ -517,10 +554,14 @@ main(int argc, char * const argv[])
 	if (workercnt == 0)
 		workercnt = mode & MODE_SUBMISSION ? 2 : get_cores();
 
+	/* disable collector for submission mode */
+	if (mode & MODE_SUBMISSION)
+		collector_interval = 0;
+
 	/* any_of failover maths need batchsize to be smaller than queuesize */
 	if (batchsize > queuesize) {
 		fprintf(stderr, "error: batchsize must be smaller than queuesize\n");
-		exit(-1);
+		exit(1);
 	}
 
 	if (relay_logfile != NULL && (mode & MODE_TEST) == 0) {
@@ -528,10 +569,14 @@ main(int argc, char * const argv[])
 		if (f == NULL) {
 			fprintf(stderr, "error: failed to open logfile '%s': %s\n",
 					relay_logfile, strerror(errno));
-			exit(-1);
+			exit(1);
 		}
 		relay_stdout = f;
 		relay_stderr = f;
+	} else if (mode & MODE_CONFIGTEST) {
+		/* suppress stdout, just show errors */
+		relay_stdout = NULL;
+		relay_stderr = stderr;
 	} else {
 		relay_stdout = stdout;
 		relay_stderr = stderr;
@@ -542,13 +587,13 @@ main(int argc, char * const argv[])
 		if (relay_logfile == NULL) {
 			fprintf(stderr,
 					"You must specify logfile if you want daemonisation!\n");
-			exit(-1);
+			exit(1);
 		}
 
 		if (mode & MODE_TEST) {
 			fprintf(stderr,
 					"You cannot use test mode if you want daemonisation!\n");
-			exit(-1);
+			exit(1);
 		}
 	}
 
@@ -557,13 +602,12 @@ main(int argc, char * const argv[])
 		if (pidfile_handle == NULL) {
 			fprintf(stderr, "failed to open pidfile '%s': %s\n",
 					pidfile, strerror(errno));
-			exit(-1);
+			exit(1);
 		}
 	}
 
 	if (mode & MODE_DAEMON) {
 		pid_t p;
-		int fds[2];
 
 		if (pipe(fds) < 0) {
 			fprintf(stderr, 
@@ -579,19 +623,19 @@ main(int argc, char * const argv[])
 			exit(1);
 		} else if (p == 0) {
 			char msg[1024];
-			/* child, for again so our intermediate process can die and
+			/* child, fork again so our intermediate process can die and
 			 * its child becomes an orphan to init */
 			p = fork();
 			if (p < 0) {
-				snprintf(msg, sizeof(msg), "failed to fork daemon process: %s",
-						strerror(errno));
+				snprintf(msg, sizeof(msg),
+						"failed to fork daemon process: %s\n", strerror(errno));
 				/* fool compiler */
 				if (write(fds[1], msg, strlen(msg) + 1) > -2)
 					exit(1); /* not that retcode matters */
 			} else if (p == 0) {
 				/* child, this is the final daemon process */
 				if (setsid() < 0) {
-					snprintf(msg, sizeof(msg), "failed to setsid(): %s",
+					snprintf(msg, sizeof(msg), "failed to setsid(): %s\n",
 							strerror(errno));
 					/* fool compiler */
 					if (write(fds[1], msg, strlen(msg) + 1) > -2)
@@ -602,11 +646,6 @@ main(int argc, char * const argv[])
 				close(0);
 				close(1);
 				close(2);
-
-				/* we're fine, flag our grandparent (= the original
-				 * process being called) it can terminate */
-				if (write(fds[1], "OK", 3) < -2)
-					*msg = '\0';  /* unreachable dummy to fool compiler */
 			} else {
 				/* parent, die */
 				exit(0);
@@ -630,63 +669,93 @@ main(int argc, char * const argv[])
 				if (len == 1024)
 					len--;
 				msg[len] = '\0';
-				fprintf(stderr, "%s\n", msg);
+				fprintf(stderr, "%s", msg);
 				exit(1);
 			}
 		}
-
-		close(fds[0]);
-		close(fds[1]);
+	} else {
+		fds[0] = fds[1] = -1;
 	}
+#define exit_err(...) { \
+	if (fds[0] != -1) { \
+		char msg[1024]; \
+		snprintf(msg, sizeof(msg), __VA_ARGS__); \
+		/* fool compiler */ \
+		if (write(fds[1], msg, strlen(msg) + 1) > -2) \
+			exit(1); /* not that retcode matters */ \
+	} else { \
+		logerr(__VA_ARGS__); \
+		exit(1); \
+	} \
+}
+#define exit_ok { \
+	if (fds[0] != -1) { \
+		/* we're fine, flag our grandparent (= the original
+		 * process being called) it can terminate */ \
+		if (write(fds[1], "OK", 3) < -2) \
+			exit(1);  /* if our grantparent died, we better do too */ \
+		close(fds[0]); \
+		close(fds[1]); \
+	} \
+	fds[0] = fds[1] = -1; \
+	startup_success = 1; \
+}
 
 	if (pidfile_handle) {
-		fprintf(pidfile_handle, "%d\n", getpid());
+		fprintf(pidfile_handle, "%zu\n", (ssize_t)getpid());
 		fclose(pidfile_handle);
 	}
 
 
 	logout("starting carbon-c-relay v%s (%s), pid=%d\n",
 			VERSION, GIT_VERSION, getpid());
-	fprintf(relay_stdout, "configuration:\n");
-	fprintf(relay_stdout, "    relay hostname = %s\n", relay_hostname);
-	fprintf(relay_stdout, "    listen port = %u\n", listenport);
-	if (listeninterface != NULL)
-		fprintf(relay_stdout, "    listen interface = %s\n", listeninterface);
-	fprintf(relay_stdout, "    workers = %d\n", workercnt);
-	fprintf(relay_stdout, "    send batch size = %d\n", batchsize);
-	fprintf(relay_stdout, "    server queue size = %d\n", queuesize);
-	fprintf(relay_stdout, "    server max stalls = %d\n", maxstalls);
-	fprintf(relay_stdout, "    statistics submission interval = %ds\n",
-			collector_interval);
-	fprintf(relay_stdout, "    listen backlog = %u\n", listenbacklog);
-	if (sockbufsize > 0)
-		fprintf(relay_stdout, "    socket bufsize = %u\n", sockbufsize);
-	fprintf(relay_stdout, "    server connection IO timeout = %dms\n",
-			iotimeout);
-	if (allowed_chars != NULL)
-		fprintf(relay_stdout, "    extra allowed characters = %s\n",
-				allowed_chars);
-	if (mode & MODE_DEBUG)
-		fprintf(relay_stdout, "    debug = true\n");
+	if (relay_stdout != NULL) {
+		fprintf(relay_stdout, "configuration:\n");
+		fprintf(relay_stdout, "    relay hostname = %s\n", relay_hostname);
+		fprintf(relay_stdout, "    workers = %d\n", workercnt);
+		fprintf(relay_stdout, "    send batch size = %d\n", batchsize);
+		fprintf(relay_stdout, "    server queue size = %d\n", queuesize);
+		fprintf(relay_stdout, "    server max stalls = %d\n", maxstalls);
+		fprintf(relay_stdout, "    listen backlog = %u\n", listenbacklog);
+		if (sockbufsize > 0)
+			fprintf(relay_stdout, "    socket bufsize = %u\n", sockbufsize);
+		fprintf(relay_stdout, "    server connection IO timeout = %dms\n",
+				iotimeout);
+		if (allowed_chars != NULL)
+			fprintf(relay_stdout, "    extra allowed characters = %s\n",
+					allowed_chars);
+		if (mode & MODE_DEBUG)
+			fprintf(relay_stdout, "    debug = true\n");
 #ifdef ENABLE_TRACE
-	if (mode & MODE_TRACE)
-		fprintf(relay_stdout, "    trace = true\n");
+		if (mode & MODE_TRACE)
+			fprintf(relay_stdout, "    trace = true\n");
 #endif
-	else if (mode & MODE_SUBMISSION)
-		fprintf(relay_stdout, "    submission = true\n");
-	else if (mode & MODE_DAEMON)
-		fprintf(relay_stdout, "    daemon = true\n");
-	fprintf(relay_stdout, "    routes configuration = %s\n", config);
-	fprintf(relay_stdout, "\n");
-
-	if ((rtr = router_readconfig(NULL, config,
-					queuesize, batchsize, maxstalls,
-					iotimeout, sockbufsize)) == NULL)
-	{
-		logerr("failed to read configuration '%s'\n", config);
-		return 1;
+		else if (mode & MODE_SUBMISSION)
+			fprintf(relay_stdout, "    submission = true\n");
+		else if (mode & MODE_DAEMON)
+			fprintf(relay_stdout, "    daemon = true\n");
+		fprintf(relay_stdout, "    configuration = %s\n", config);
+		fprintf(relay_stdout, "\n");
 	}
-	router_optimise(rtr);
+
+#ifdef HAVE_SSL
+	/* this needs to happen once, globally */
+	SSL_library_init();
+	SSL_load_error_strings();
+	OpenSSL_add_all_algorithms();
+#endif
+
+	if ((rtr = router_readconfig(NULL, config, workercnt,
+					queuesize, batchsize, maxstalls,
+					iotimeout, sockbufsize, listenport)) == NULL)
+	{
+		exit_err("failed to read configuration '%s'\n", config);
+	}
+	router_optimise(rtr, optimiserthreshold);
+
+	/* we're done if all we wanted was to test the config */
+	if (mode & MODE_CONFIGTEST)
+		exit(0);
 
 	aggrs = router_getaggregators(rtr);
 	numaggregators = aggregator_numaggregators(aggrs);
@@ -699,7 +768,8 @@ main(int argc, char * const argv[])
 	} else {
 		fprintf(relay_stdout, "parsed configuration follows:\n");
 		router_printconfig(rtr, relay_stdout,
-				PMODE_AGGR + (mode & MODE_DEBUG ? PMODE_DEBUG : PMODE_NORM));
+				PMODE_AGGR + (mode & MODE_DEBUG ?
+					PMODE_DEBUG | PMODE_PEMT : PMODE_NORM));
 	}
 	fprintf(relay_stdout, "\n");
 
@@ -712,6 +782,11 @@ main(int argc, char * const argv[])
 		while (fgets(metricbuf, sizeof(metricbuf), stdin) != NULL) {
 			if ((p = strchr(metricbuf, '\n')) != NULL)
 				*p = '\0';
+			p = metricbuf;
+			while (isspace(*p))
+				p++;
+			if (*p == '\0')
+				continue;
 			router_test(rtr, metricbuf);
 		}
 
@@ -719,68 +794,58 @@ main(int argc, char * const argv[])
 	}
 
 	if (signal(SIGINT, sig_handler) == SIG_ERR) {
-		logerr("failed to create SIGINT handler: %s\n", strerror(errno));
-		return 1;
+		exit_err("failed to create SIGINT handler: %s\n", strerror(errno));
 	}
 	if (signal(SIGTERM, sig_handler) == SIG_ERR) {
-		logerr("failed to create SIGTERM handler: %s\n", strerror(errno));
-		return 1;
+		exit_err("failed to create SIGTERM handler: %s\n", strerror(errno));
 	}
 	if (signal(SIGQUIT, sig_handler) == SIG_ERR) {
-		logerr("failed to create SIGQUIT handler: %s\n", strerror(errno));
-		return 1;
+		exit_err("failed to create SIGQUIT handler: %s\n", strerror(errno));
 	}
 	if (signal(SIGHUP, sig_handler) == SIG_ERR) {
-		logerr("failed to create SIGHUP handler: %s\n", strerror(errno));
-		return 1;
+		exit_err("failed to create SIGHUP handler: %s\n", strerror(errno));
 	}
 	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
-		logerr("failed to ignore SIGPIPE: %s\n", strerror(errno));
-		return 1;
+		exit_err("failed to ignore SIGPIPE: %s\n", strerror(errno));
 	}
 
-	workers = malloc(sizeof(dispatcher *) * (1 + workercnt + 1));
+	workers = malloc(sizeof(dispatcher *) *
+			(1/*lsnr*/ + workercnt + 1/*sentinel*/));
 	if (workers == NULL) {
-		logerr("failed to allocate memory for workers\n");
-		return 1;
+		exit_err("failed to allocate memory for workers\n");
 	}
 
-	if (bindlisten(stream_sock, &stream_socklen,
-				dgram_sock, &dgram_socklen,
-				listeninterface, listenport, listenbacklog) < 0) {
-		logerr("failed to bind on port %s:%d: %s\n",
-				listeninterface == NULL ? "" : listeninterface,
-				listenport, strerror(errno));
-		return -1;
-	}
 	dispatch_set_bufsize(sockbufsize);
-	for (ch = 0; ch < stream_socklen; ch++) {
-		if (dispatch_addlistener(stream_sock[ch]) != 0) {
-			logerr("failed to add listener\n");
-			return -1;
+	if (dispatch_init_listeners() != 0) {
+		exit_err("failed to allocate listeners\n");
+	}
+
+	lsnrs = router_get_listeners(rtr);
+	for ( ; lsnrs != NULL; lsnrs = lsnrs->next) {
+		if (bindlisten(lsnrs, listenbacklog) != 0) {
+			exit_err("failed to setup listener\n");
+		}
+		if (dispatch_addlistener(lsnrs) != 0) {
+			exit_err("failed to add listener\n");
 		}
 	}
-	for (ch = 0; ch < dgram_socklen; ch++) {
-		if (dispatch_addlistener_udp(dgram_sock[ch]) != 0) {
-			logerr("failed to listen to datagram socket\n");
-			return -1;
-		}
-	}
-	if ((workers[0] = dispatch_new_listener()) == NULL)
-		logerr("failed to add listener\n");
+	/* ensure the listener id is at the end for regex_t array hack */
+	if ((workers[0] = dispatch_new_listener((unsigned char)workercnt)) == NULL)
+		logerr("failed to add listener dispatcher\n");
 
 	if (allowed_chars == NULL)
 		allowed_chars = "-_:#";
 	logout("starting %d workers\n", workercnt);
-	for (id = 1; id < 1 + workercnt; id++) {
-		workers[id + 0] = dispatch_new_connection(rtr, allowed_chars);
-		if (workers[id + 0] == NULL) {
+	for (id = 0; id < workercnt; id++) {
+		workers[id + 1] = dispatch_new_connection(
+				(unsigned char)id, rtr, allowed_chars);
+		if (workers[id + 1] == NULL) {
 			logerr("failed to add worker %d\n", id);
 			break;
 		}
 	}
-	workers[id + 0] = NULL;
-	if (id < 1 + workercnt) {
+	workers[id + 1] = NULL;  /* sentinel */
+	if (id < workercnt) {
 		logerr("shutting down due to errors\n");
 		keep_running = 0;
 	}
@@ -788,7 +853,8 @@ main(int argc, char * const argv[])
 	/* server used for delivering metrics produced inside the relay,
 	 * that is, the collector (statistics) */
 	if ((internal_submission = server_new(
-					"internal", listenport, CON_PIPE, NULL, 3000,
+					"internal", listenport, T_LINEMODE, W_PLAIN, CON_PIPE,
+					NULL, NULL, 3000,
 					batchsize, maxstalls, iotimeout, sockbufsize)) == NULL)
 	{
 		logerr("failed to create internal submission queue, shutting down\n");
@@ -819,8 +885,10 @@ main(int argc, char * const argv[])
 		keep_running = 0;
 	}
 
-	if (keep_running == 0)
+	if (keep_running != 0) {
 		logout("startup sequence complete\n");
+		exit_ok;
+	}
 
 	/* workers do the work, just wait */
 	while (keep_running) {
@@ -833,10 +901,11 @@ main(int argc, char * const argv[])
 
 	logout("shutting down...\n");
 	/* make sure we don't accept anything new anymore */
-	for (ch = 0; ch < stream_socklen; ch++)
-		dispatch_removelistener(stream_sock[ch]);
-	destroy_usock(listenport);
-	logout("closed listeners for port %u\n", listenport);
+	lsnrs = router_get_listeners(rtr);
+	for ( ; lsnrs != NULL; lsnrs = lsnrs->next) {
+		dispatch_removelistener(lsnrs);
+		shutdownclose(lsnrs);
+	}
 	/* since workers will be freed, stop querying the structures */
 	collector_stop();
 	server_shutdown(internal_submission);
@@ -872,5 +941,12 @@ main(int argc, char * const argv[])
 
 	logout("stopped carbon-c-relay v%s (%s)\n", VERSION, GIT_VERSION);
 
+	/* if we never called exit_ok before, it means we shut down due to
+	 * errors, so exit with an error */
+	if (startup_success == 0) {
+		/* this message will never reach the log/screen when not
+		 * daemonised, because the logger already shut down */
+		exit_err("startup failed, review the logs for details\n");
+	}
 	return 0;
 }
