@@ -18,7 +18,9 @@
 #ifndef HAVE_RELAY_H
 #define HAVE_RELAY_H 1
 
-#define VERSION "2.5"
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
 #define METRIC_BUFSIZ 8192
 
@@ -28,22 +30,38 @@
 #define MODE_TEST       (1 << 2)
 #define MODE_DAEMON     (1 << 3)
 #define MODE_TRACE      (1 << 4)
+#define MODE_CONFIGTEST (1 << 5)
 extern unsigned char mode;
 
 #ifdef ENABLE_TRACE
-#define tracef(...) if (mode & MODE_TRACE) fprintf(stdout, __VA_ARGS__)
+#define tracef(...) if (mode & MODE_TRACE) { fprintf(stdout, __VA_ARGS__); fflush(stdout); }
 #else
 #define tracef(...) /* noop */
 #endif
 
-typedef enum { CON_TCP, CON_UDP, CON_PIPE, CON_FILE } serv_ctype;
-
 extern char relay_hostname[];
 
 enum logdst { LOGOUT, LOGERR };
 
+typedef enum { CON_TCP, CON_UDP, CON_PIPE, CON_FILE, CON_UNIX } con_proto;
+typedef enum { T_LINEMODE } con_type;
+typedef enum { W_PLAIN, W_GZIP, W_LZ4, W_SSL } con_trnsp;
+
 int relaylog(enum logdst dest, const char *fmt, ...);
 #define logout(args...) relaylog(LOGOUT, args)
 #define logerr(args...) relaylog(LOGERR, args)
 
+#define saddr_ntop(SADDR, BUF) { \
+	BUF[0] = '\0'; \
+	if ((SADDR)->ai_family == PF_INET) { \
+		inet_ntop((SADDR)->ai_family, \
+					&((struct sockaddr_in *)(SADDR)->ai_addr)->sin_addr, \
+					BUF, sizeof(BUF)); \
+	} else if ((SADDR)->ai_family == PF_INET6) { \
+		inet_ntop((SADDR)->ai_family, \
+					&((struct sockaddr_in6 *)(SADDR)->ai_addr)->sin6_addr, \
+					BUF, sizeof(BUF)); \
+	} \
+}
+
 #endif
