@@ -27,6 +27,7 @@
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <netdb.h>
+#include <netinet/in.h>
 #include <netinet/tcp.h>
 #include <arpa/inet.h>
 #include <sys/resource.h>
@@ -38,27 +39,50 @@
 #include "collector.h"
 #include "server.h"
 
+#ifdef HAVE_GZIP
+#include <zlib.h>
+#endif
+#ifdef HAVE_LZ4
+#include <lz4.h>
+#endif
+#ifdef HAVE_SSL
+#include <openssl/ssl.h>
+#include <openssl/err.h>
+#endif
+
 struct _server {
 	const char *ip;
 	unsigned short port;
 	char *instance;
 	struct addrinfo *saddr;
+	struct addrinfo *hint;
+	char reresolve:1;
 	int fd;
+	void *strm;
+	ssize_t (*strmwrite)(void *, const void *, size_t);  /* write impl */
+	int (*strmflush)(void *);                            /* flush impl */
+	int (*strmclose)(void *);                            /* close impl */
+	const char *(*strmerror)(void *, int);               /* get last err str */
+#ifdef HAVE_SSL
+	SSL_CTX *ctx;
+#endif
 	queue *queue;
 	size_t bsize;
 	short iotimeout;
 	unsigned int sockbufsize;
 	unsigned char maxstalls:SERVER_STALL_BITS;
 	const char **batch;
-	serv_ctype ctype;
+	con_type type;
+	con_trnsp transport;
+	con_proto ctype;
 	pthread_t tid;
 	struct _server **secondaries;
 	size_t secondariescnt;
 	char failover:1;
 	char failure;       /* full byte for atomic access */
-	char running:1;
+	char running;       /* full byte for atomic access */
 	char keep_running;  /* full byte for atomic access */
-	unsigned char stallseq:SERVER_STALL_BITS;
+	unsigned char stallseq;  /* full byte for atomic access */
 	size_t metrics;
 	size_t dropped;
 	size_t stalls;
@@ -70,6 +94,140 @@ struct _server {
 };
 
 
+/* connection specific writers and closers */
+
+typedef struct _z_strm {
+	char obuf[METRIC_BUFSIZ];
+	int sock;
+	union {
+#ifdef HAVE_LZ4
+		LZ4_stream_t *lz;
+#endif
+		void *dummy;
+	} hdl;
+} z_strm;
+
+/* ordinary socket */
+static inline ssize_t
+sockwrite(void *strm, const void *buf, size_t sze)
+{
+	return write(*((int *)strm), buf, sze);
+}
+
+static inline int
+sockflush(void *strm)
+{
+	/* noop, we don't use a stream in the normal case */
+	return 0;
+}
+
+static inline int
+sockclose(void *strm)
+{
+	return close(*((int *)strm));
+}
+
+static inline const char *
+sockerror(void *strm, int rval)
+{
+	(void)strm;
+	(void)rval;
+	return strerror(errno);
+}
+
+#ifdef HAVE_GZIP
+/* gzip wrapped socket */
+static inline ssize_t
+gzipwrite(void *strm, const void *buf, size_t sze)
+{
+	return (ssize_t)gzwrite((gzFile)strm, buf, (unsigned)sze);
+}
+
+static inline int
+gzipflush(void *strm)
+{
+	return gzflush((gzFile)strm, Z_SYNC_FLUSH);
+}
+
+static inline int
+gzipclose(void *strm)
+{
+	return gzclose((gzFile)strm);
+}
+#endif
+
+#ifdef HAVE_LZ4
+/* lz4 wrapped socket */
+static inline ssize_t
+lzwrite(void *strm, const void *buf, size_t sze)
+{
+	int oret;
+	char *obuf = ((z_strm *)strm)->obuf;
+	int cret;
+
+	cret = LZ4_compress_fast_continue(((z_strm *)strm)->hdl.lz,
+			buf, obuf, sze, METRIC_BUFSIZ, 0);
+	if (cret == 0)
+		return -1;  /* we must reset/free lz */
+	while (cret > 0) {
+		oret = write(((z_strm *)strm)->sock, obuf, cret);
+		if (oret < 0)
+			return -1;  /* failure is failure */
+		/* update counters to possibly retry the remaining bit */
+		obuf += oret;
+		cret -= oret;
+	}
+
+	return sze;
+}
+
+static inline int
+lzflush(void *strm)
+{
+	return 0;
+}
+
+static inline int
+lzclose(void *strm)
+{
+	int ret = close(((z_strm *)strm)->sock);
+	LZ4_freeStream(((z_strm *)strm)->hdl.lz);
+	free(strm);
+	return ret;
+}
+#endif
+
+#ifdef HAVE_SSL
+/* (Open|Libre)SSL wrapped socket */
+static inline ssize_t
+sslwrite(void *strm, const void *buf, size_t sze)
+{
+	return (ssize_t)SSL_write((SSL *)strm, buf, (int)sze);
+}
+
+static inline int
+sslflush(void *strm)
+{
+	/* noop */
+	return 0;
+}
+
+static inline int
+sslclose(void *strm)
+{
+	int sock = SSL_get_fd((SSL *)strm);
+	SSL_free((SSL *)strm);
+	return close(sock);
+}
+
+static inline const char *
+sslerror(void *strm, int rval)
+{
+	int err = SSL_get_error((SSL *)strm, rval);
+	return ERR_reason_error_string(err);
+}
+#endif
+
 /**
  * Reads from the queue and sends items to the remote server.  This
  * function is designed to be a thread.  Data sending is attempted to be
@@ -108,9 +266,14 @@ server_queuereader(void *d)
 			if (self->ctype == CON_TCP && self->fd >= 0 &&
 					idle++ > DISCONNECT_WAIT_TIME)
 			{
-				close(self->fd);
+				self->strmclose(self->strm);
 				self->fd = -1;
 			}
+			if (idle == 1)
+				/* ensure blocks are pushed out as soon as we're idling,
+				 * this allows compressors to benefit from a larger
+				 * stream of data to gain better compresion */
+				self->strmflush(self->strm);
 			gettimeofday(&stop, NULL);
 			__sync_add_and_fetch(&(self->ticks), timediff(start, stop));
 			if (__sync_bool_compare_and_swap(&(self->keep_running), 0, 0))
@@ -235,6 +398,26 @@ server_queuereader(void *d)
 
 		/* try to connect */
 		if (self->fd < 0) {
+			if (self->reresolve) {  /* can only be CON_UDP/CON_TCP */
+				struct addrinfo *saddr;
+				char sport[8];
+
+				/* re-lookup the address info, if it fails, stay with
+				 * whatever we have such that resolution errors incurred
+				 * after starting the relay won't make it fail */
+				freeaddrinfo(self->saddr);
+				snprintf(sport, sizeof(sport), "%u", self->port);
+				if (getaddrinfo(self->ip, sport, self->hint, &saddr) == 0) {
+					self->saddr = saddr;
+				} else {
+					if (__sync_fetch_and_add(&(self->failure), 1) == 0)
+						logerr("failed to resolve %s:%u, server unavailable\n",
+								self->ip, self->port);
+					self->saddr = NULL;
+					/* this will break out below */
+				}
+			}
+
 			if (self->ctype == CON_PIPE) {
 				int intconn[2];
 				if (pipe(intconn) < 0) {
@@ -242,28 +425,8 @@ server_queuereader(void *d)
 						logerr("failed to create pipe: %s\n", strerror(errno));
 					continue;
 				}
-				dispatch_addconnection(intconn[0]);
+				dispatch_addconnection(intconn[0], NULL);
 				self->fd = intconn[1];
-			} else if (self->ctype == CON_UDP) {
-				if ((self->fd = socket(self->saddr->ai_family,
-								self->saddr->ai_socktype,
-								self->saddr->ai_protocol)) < 0)
-				{
-					if (__sync_fetch_and_add(&(self->failure), 1) == 0)
-						logerr("failed to create udp socket: %s\n",
-								strerror(errno));
-					continue;
-				}
-				if (connect(self->fd,
-						self->saddr->ai_addr, self->saddr->ai_addrlen) < 0)
-				{
-					if (__sync_fetch_and_add(&(self->failure), 1) == 0)
-						logerr("failed to connect udp socket: %s\n",
-								strerror(errno));
-					close(self->fd);
-					self->fd = -1;
-					continue;
-				}
 			} else if (self->ctype == CON_FILE) {
 				if ((self->fd = open(self->ip,
 								O_WRONLY | O_APPEND | O_CREAT,
@@ -274,92 +437,161 @@ server_queuereader(void *d)
 								self->ip, strerror(errno));
 					continue;
 				}
-			} else {
+			} else if (self->ctype == CON_UDP) {
+				struct addrinfo *walk;
+
+				for (walk = self->saddr; walk != NULL; walk = walk->ai_next) {
+					if ((self->fd = socket(walk->ai_family,
+									walk->ai_socktype,
+									walk->ai_protocol)) < 0)
+					{
+						if (walk->ai_next == NULL &&
+								__sync_fetch_and_add(&(self->failure), 1) == 0)
+							logerr("failed to create udp socket: %s\n",
+									strerror(errno));
+						continue;
+					}
+					if (connect(self->fd, walk->ai_addr, walk->ai_addrlen) < 0)
+					{
+						if (walk->ai_next == NULL &&
+								__sync_fetch_and_add(&(self->failure), 1) == 0)
+							logerr("failed to connect udp socket: %s\n",
+									strerror(errno));
+						close(self->fd);
+						self->fd = -1;
+						continue;
+					}
+
+					/* we made it up here, so this connection is usable */
+					break;
+				}
+				/* if this didn't resolve to anything, treat as failure */
+				if (self->saddr == NULL)
+					__sync_add_and_fetch(&(self->failure), 1);
+				/* if all addrinfos failed, try again later */
+				if (self->fd < 0)
+					continue;
+			} else {  /* CON_TCP */
 				int ret;
 				int args;
+				struct addrinfo *walk;
 
-				if ((self->fd = socket(self->saddr->ai_family,
-								self->saddr->ai_socktype,
-								self->saddr->ai_protocol)) < 0)
-				{
-					if (__sync_fetch_and_add(&(self->failure), 1) == 0)
-						logerr("failed to create socket: %s\n",
+				for (walk = self->saddr; walk != NULL; walk = walk->ai_next) {
+					if ((self->fd = socket(walk->ai_family,
+									walk->ai_socktype,
+									walk->ai_protocol)) < 0)
+					{
+						if (walk->ai_next == NULL &&
+								__sync_fetch_and_add(&(self->failure), 1) == 0)
+							logerr("failed to create socket: %s\n",
+									strerror(errno));
+						continue;
+					}
+
+					/* put socket in non-blocking mode such that we can
+					 * poll() (time-out) on the connect() call */
+					args = fcntl(self->fd, F_GETFL, NULL);
+					if (fcntl(self->fd, F_SETFL, args | O_NONBLOCK) < 0) {
+						logerr("failed to set socket non-blocking mode: %s\n",
 								strerror(errno));
-					continue;
-				}
-
-				/* put socket in non-blocking mode such that we can
-				 * poll() (time-out) on the connect() call */
-				args = fcntl(self->fd, F_GETFL, NULL);
-				if (fcntl(self->fd, F_SETFL, args | O_NONBLOCK) < 0) {
-					logerr("failed to set socket non-blocking mode: %s\n",
-							strerror(errno));
-					close(self->fd);
-					self->fd = -1;
-					continue;
-				}
-				ret = connect(self->fd,
-						self->saddr->ai_addr, self->saddr->ai_addrlen);
-
-				if (ret < 0 && errno == EINPROGRESS) {
-					/* wait for connection to succeed if the OS thinks
-					 * it can succeed */
-					struct pollfd ufds[1];
-					ufds[0].fd = self->fd;
-					ufds[0].events = POLLIN | POLLOUT;
-					ret = poll(ufds, 1, self->iotimeout + (rand() % 100));
-					if (ret == 0) {
-						/* time limit expired */
-						if (__sync_fetch_and_add(&(self->failure), 1) == 0)
-							logerr("failed to connect() to "
-									"%s:%u: Operation timed out\n",
-									self->ip, self->port);
 						close(self->fd);
 						self->fd = -1;
 						continue;
-					} else if (ret < 0) {
-						/* some select error occurred */
-						if (__sync_fetch_and_add(&(self->failure), 1) == 0)
-							logerr("failed to poll() for %s:%u: %s\n",
-									self->ip, self->port, strerror(errno));
-						close(self->fd);
-						self->fd = -1;
-						continue;
-					} else {
-						if (ufds[0].revents & POLLHUP) {
-							if (__sync_fetch_and_add(&(self->failure), 1) == 0)
-								logerr("failed to connect() for %s:%u: "
-										"Hangup\n", self->ip, self->port);
+					}
+					ret = connect(self->fd, walk->ai_addr, walk->ai_addrlen);
+
+					if (ret < 0 && errno == EINPROGRESS) {
+						/* wait for connection to succeed if the OS thinks
+						 * it can succeed */
+						struct pollfd ufds[1];
+						ufds[0].fd = self->fd;
+						ufds[0].events = POLLIN | POLLOUT;
+						ret = poll(ufds, 1, self->iotimeout + (rand() % 100));
+						if (ret == 0) {
+							/* time limit expired */
+							if (walk->ai_next == NULL &&
+									__sync_fetch_and_add(
+										&(self->failure), 1) == 0)
+								logerr("failed to connect() to "
+										"%s:%u: Operation timed out\n",
+										self->ip, self->port);
 							close(self->fd);
 							self->fd = -1;
 							continue;
+						} else if (ret < 0) {
+							/* some select error occurred */
+							if (walk->ai_next &&
+									__sync_fetch_and_add(
+										&(self->failure), 1) == 0)
+								logerr("failed to poll() for %s:%u: %s\n",
+										self->ip, self->port, strerror(errno));
+							close(self->fd);
+							self->fd = -1;
+							continue;
+						} else {
+							if (ufds[0].revents & POLLHUP) {
+								if (walk->ai_next == NULL &&
+										__sync_fetch_and_add(
+											&(self->failure), 1) == 0)
+									logerr("failed to connect() for %s:%u: "
+											"Connection refused\n",
+											self->ip, self->port);
+								close(self->fd);
+								self->fd = -1;
+								continue;
+							}
 						}
+					} else if (ret < 0) {
+						if (walk->ai_next == NULL &&
+								__sync_fetch_and_add(&(self->failure), 1) == 0)
+						{
+							logerr("failed to connect() to %s:%u: %s\n",
+									self->ip, self->port, strerror(errno));
+							dispatch_check_rlimit_and_warn();
+						}
+						close(self->fd);
+						self->fd = -1;
+						continue;
 					}
-				} else if (ret < 0) {
-					if (__sync_fetch_and_add(&(self->failure), 1) == 0) {
-						logerr("failed to connect() to %s:%u: %s\n",
-								self->ip, self->port, strerror(errno));
-						dispatch_check_rlimit_and_warn();
+
+					/* make socket blocking again */
+					if (fcntl(self->fd, F_SETFL, args) < 0) {
+						logerr("failed to remove socket non-blocking "
+								"mode: %s\n", strerror(errno));
+						close(self->fd);
+						self->fd = -1;
+						continue;
 					}
-					close(self->fd);
-					self->fd = -1;
-					continue;
-				}
 
-				/* make socket blocking again */
-				if (fcntl(self->fd, F_SETFL, args) < 0) {
-					logerr("failed to remove socket non-blocking mode: %s\n",
-							strerror(errno));
-					close(self->fd);
-					self->fd = -1;
-					continue;
-				}
+					/* disable Nagle's algorithm, issue #208 */
+					args = 1;
+					if (setsockopt(self->fd, IPPROTO_TCP, TCP_NODELAY,
+								&args, sizeof(args)) != 0)
+						; /* ignore */
 
-				/* disable Nagle's algorithm, issue #208 */
-				args = 1;
-				if (setsockopt(self->fd, IPPROTO_TCP, TCP_NODELAY,
-							&args, sizeof(args)) != 0)
-					; /* ignore */
+#ifdef TCP_USER_TIMEOUT
+					/* break out of connections when no ACK is being
+					 * received for +- 10 seconds instead of
+					 * retransmitting for +- 15 minutes available on
+					 * linux >= 2.6.37
+					 * the 10 seconds is in line with the SO_SNDTIMEO
+					 * set on the socket below */
+					args = 10000 + (rand() % 300);
+					if (setsockopt(self->fd, IPPROTO_TCP, TCP_USER_TIMEOUT,
+								&args, sizeof(args)) != 0)
+						; /* ignore */
+#endif
+
+					/* if we reached up here, we're good to go, so don't
+					 * continue with the other addrinfos */
+					break;
+				}
+				/* if this didn't resolve to anything, treat as failure */
+				if (self->saddr == NULL)
+					__sync_add_and_fetch(&(self->failure), 1);
+				/* all available addrinfos failed on us */
+				if (self->fd < 0)
+					continue;
 			}
 
 			/* ensure we will break out of connections being stuck more
@@ -367,17 +599,75 @@ server_queuereader(void *d)
 			timeout.tv_sec = 10;
 			timeout.tv_usec = (rand() % 300) * 1000;
 			if (setsockopt(self->fd, SOL_SOCKET, SO_SNDTIMEO,
-					&timeout, sizeof(timeout)) != 0)
+						&timeout, sizeof(timeout)) != 0)
 				; /* ignore */
 			if (self->sockbufsize > 0)
 				if (setsockopt(self->fd, SOL_SOCKET, SO_SNDBUF,
-						&self->sockbufsize, sizeof(self->sockbufsize)) != 0)
+							&self->sockbufsize, sizeof(self->sockbufsize)) != 0)
 					; /* ignore */
 #ifdef SO_NOSIGPIPE
-			if (self->ctype == CON_TCP || self->ctype == CON_UDP)
-				if (setsockopt(self->fd, SOL_SOCKET, SO_NOSIGPIPE, NULL, 0) != 0)
+			if (self->ctype == CON_TCP || self->ctype == CON_UDP) {
+				int enable = 1;
+				if (setsockopt(self->fd, SOL_SOCKET, SO_NOSIGPIPE,
+							(void *)&enable, sizeof(enable)) != 0)
 					logout("warning: failed to ignore SIGPIPE on socket: %s\n",
 							strerror(errno));
+			}
+#endif
+
+#ifdef HAVE_GZIP
+			if (self->transport == W_GZIP) {
+				self->strm = gzdopen(self->fd, "w");
+				if (self->strm == Z_NULL) {
+					logerr("failed to open gzip stream: %s\n", strerror(errno));
+					close(self->fd);
+					self->fd = -1;
+					continue;
+				}
+			}
+#endif
+#ifdef HAVE_LZ4
+			if (self->transport == W_LZ4) {
+				z_strm *s = (z_strm *)malloc(sizeof(z_strm));
+				if (s == NULL) {
+					logerr("failed to allocate lz4 stream: %s\n",
+							strerror(errno));
+					close(self->fd);
+					self->fd = -1;
+					continue;
+				}
+				s->sock = self->fd;
+				s->hdl.lz = LZ4_createStream();
+				if (s->hdl.lz == NULL) {
+					logerr("failed to create lz4 stream: out of memory\n");
+					free(s);
+					close(self->fd);
+					self->fd = -1;
+					continue;
+				}
+				self->strm = s;
+			}
+#endif
+#ifdef HAVE_SSL
+			if (self->transport == W_SSL) {
+				int rv;
+				self->strm = SSL_new(self->ctx);
+				if (SSL_set_fd(self->strm, self->fd) == 0) {
+					logerr("failed to SSL_set_fd: %s\n",
+							ERR_reason_error_string(ERR_get_error()));
+					self->strmclose(self->strm);
+					self->fd = -1;
+					continue;
+				}
+				if ((rv = SSL_connect(self->strm)) != 1) {
+					logerr("failed to  connect ssl stream: %s\n",
+							ERR_reason_error_string(
+								SSL_get_error(self->strm, rv)));
+					self->strmclose(self->strm);
+					self->fd = -1;
+					continue;
+				}
+			}
 #endif
 		}
 
@@ -413,14 +703,19 @@ server_queuereader(void *d)
 			 * resuming to complete.  So, use a loop, but to avoid
 			 * getting endlessly stuck on this, only try a limited
 			 * number of times for a single metric. */
-			for (cnt = 0, p = *metric;
-					cnt < 10 &&
-						(slen = write(self->fd, p, len)) != len &&
-						slen >= 0;
-					cnt++)
-			{
-				p += slen;
-				len -= slen;
+			for (cnt = 0, p = *metric; cnt < 10; cnt++) {
+				if ((slen = self->strmwrite(self->strm, p, len)) != len) {
+					if (slen >= 0) {
+						p += slen;
+						len -= slen;
+					} else if (errno != EINTR) {
+						break;
+					}
+					/* allow the remote to catch up */
+					usleep((50 + (rand() % 150)) * 1000);  /* 50ms - 200ms */
+				} else {
+					break;
+				}
 			}
 			if (slen != len) {
 				/* not fully sent (after tries), or failure
@@ -430,8 +725,9 @@ server_queuereader(void *d)
 						__sync_fetch_and_add(&(self->failure), 1) == 0)
 					logerr("failed to write() to %s:%u: %s\n",
 							self->ip, self->port,
-							(slen < 0 ? strerror(errno) : "incomplete write"));
-				close(self->fd);
+							(slen < 0 ? self->strmerror(self->strm, slen) :
+							 "incomplete write"));
+				self->strmclose(self->strm);
 				self->fd = -1;
 				/* put back stuff we couldn't process */
 				for (; *metric != NULL; metric++) {
@@ -458,10 +754,10 @@ server_queuereader(void *d)
 
 		idle = 0;
 	}
-	self->running = 0;
+	__sync_and_and_fetch(&(self->running), 0);
 
 	if (self->fd >= 0)
-		close(self->fd);
+		self->strmclose(self->strm);
 	if (secpos != NULL)
 		free(secpos);
 	return NULL;
@@ -476,8 +772,11 @@ server *
 server_new(
 		const char *ip,
 		unsigned short port,
-		serv_ctype ctype,
+		con_type type,
+		con_trnsp transport,
+		con_proto ctype,
 		struct addrinfo *saddr,
+		struct addrinfo *hint,
 		size_t qsize,
 		size_t bsize,
 		int maxstalls,
@@ -489,6 +788,8 @@ server_new(
 	if ((ret = malloc(sizeof(server))) == NULL)
 		return NULL;
 
+	ret->type = type;
+	ret->transport = transport;
 	ret->ctype = ctype;
 	ret->tid = 0;
 	ret->secondaries = NULL;
@@ -510,9 +811,65 @@ server_new(
 		return NULL;
 	}
 	ret->fd = -1;
+	if (transport == W_PLAIN) {
+		ret->strm = &(ret->fd);
+		ret->strmwrite = &sockwrite;
+		ret->strmflush = &sockflush;
+		ret->strmclose = &sockclose;
+		ret->strmerror = &sockerror;
+	}
+#ifdef HAVE_GZIP
+	else if (transport == W_GZIP) {
+		ret->strmwrite = &gzipwrite;
+		ret->strmflush = &gzipflush;
+		ret->strmclose = &gzipclose;
+		ret->strmerror = &sockerror;
+	}
+#endif
+#ifdef HAVE_LZ4
+	else if (transport == W_LZ4) {
+		ret->strmwrite = &lzwrite;
+		ret->strmflush = &lzflush;
+		ret->strmclose = &lzclose;
+		ret->strmerror = &sockerror;
+	}
+#endif
+#ifdef HAVE_SSL
+	else if (transport == W_SSL) {
+		/* create a auto-negotiate context */
+		const SSL_METHOD *m = SSLv23_client_method();
+		ret->ctx = SSL_CTX_new(m);
+		
+		if (ret->ctx == NULL) {
+			char *err = ERR_error_string(ERR_get_error(), NULL);
+			logerr("failed to create SSL context for server "
+					"%s:%d: %s\n", ret->ip, ret->port, err);
+			free((char *)ret->ip);
+			free(ret);
+			return NULL;
+		}
+
+		ret->strmwrite = &sslwrite;
+		ret->strmflush = &sslflush;
+		ret->strmclose = &sslclose;
+		ret->strmerror = &sslerror;
+	}
+#endif
+	else {
+		logerr("no transport type defined for server!!! (this is a BUG)\n");
+		assert(0);
+	}
 	ret->saddr = saddr;
+	ret->reresolve = 0;
+	ret->hint = NULL;
+	if (hint != NULL) {
+		ret->reresolve = 1;
+		ret->hint = hint;
+	}
 	ret->queue = queue_new(qsize);
 	if (ret->queue == NULL) {
+		if (ret->hint)
+			free(ret->hint);
 		free(ret->batch);
 		free((char *)ret->ip);
 		free(ret);
@@ -537,6 +894,46 @@ server_new(
 	return ret;
 }
 
+/**
+ * Compare server s against the address info in saddr.  A server is
+ * considered to be equal is it is of the same socket family, type and
+ * protocol, and if the target address and port are the same.  When
+ * saddr is NULL, a match against the given ip is attempted, e.g. for
+ * file destinations.
+ */
+char
+server_cmp(server *s, struct addrinfo *saddr, const char *ip)
+{
+	if ((saddr == NULL || s->saddr == NULL)) {
+		if (strcmp(s->ip, ip) == 0)
+			return 0;
+	} else if (
+			s->saddr->ai_family == saddr->ai_family &&
+			s->saddr->ai_socktype == saddr->ai_socktype &&
+			s->saddr->ai_protocol == saddr->ai_protocol
+	   )
+	{
+		if (saddr->ai_family == AF_INET) {
+			struct sockaddr_in *l = ((struct sockaddr_in *)s->saddr->ai_addr);
+			struct sockaddr_in *r = ((struct sockaddr_in *)saddr->ai_addr);
+			if (l->sin_port == r->sin_port &&
+					l->sin_addr.s_addr == r->sin_addr.s_addr)
+				return 0;
+		} else if (saddr->ai_family == AF_INET6) {
+			struct sockaddr_in6 *l =
+				((struct sockaddr_in6 *)s->saddr->ai_addr);
+			struct sockaddr_in6 *r =
+				((struct sockaddr_in6 *)saddr->ai_addr);
+			if (l->sin6_port == r->sin6_port &&
+					memcmp(l->sin6_addr.s6_addr, r->sin6_addr.s6_addr,
+						sizeof(l->sin6_addr.s6_addr)) == 0)
+				return 0;
+		}
+	}
+
+	return 1;  /* not equal */
+}
+
 /**
  * Starts a previously created server using server_new().  Returns
  * errno if starting a thread failed, after which the caller should
@@ -603,16 +1000,18 @@ server_send(server *s, const char *d, char force)
 				}
 			}
 		}
-		if (failure || force || s->stallseq == s->maxstalls) {
+		if (failure || force ||
+				__sync_add_and_fetch(&(s->stallseq), 0) == s->maxstalls)
+		{
 			__sync_add_and_fetch(&(s->dropped), 1);
 			/* excess event will be dropped by the enqueue below */
 		} else {
-			s->stallseq++;
+			__sync_add_and_fetch(&(s->stallseq), 1);
 			__sync_add_and_fetch(&(s->stalls), 1);
 			return 0;
 		}
 	} else {
-		s->stallseq = 0;
+		__sync_and_and_fetch(&(s->stallseq), 0);
 	}
 	queue_enqueue(s->queue, d);
 
@@ -643,7 +1042,7 @@ server_shutdown(server *s)
 			for (i = 0; i < s->secondariescnt; i++) {
 				if (__sync_add_and_fetch(&(s->secondaries[i]->failure), 0))
 					failures++;
-				if (s->secondaries[i]->running)
+				if (__sync_add_and_fetch(&(s->secondaries[i]->running), 0))
 					inqueue += queue_len(s->secondaries[i]->queue);
 			}
 			/* loop until we all failed, or nothing is in the queues */
@@ -660,6 +1059,12 @@ server_shutdown(server *s)
 		if (inqueue != 0)
 			for (i = 0; i < s->secondariescnt; i++)
 				__sync_add_and_fetch(&(s->secondaries[i]->failure), 1);
+		/* wait for the secondaries to be stopped so we surely don't get
+		 * invalid reads when server_free is called */
+		for (i = 0; i < s->secondariescnt; i++) {
+			while (__sync_add_and_fetch(&(s->secondaries[i]->running), 0))
+				usleep((200 + (rand() % 100)) * 1000);
+		}
 	}
 
 	__sync_bool_compare_and_swap(&(s->keep_running), 1, 0);
@@ -691,6 +1096,8 @@ server_free(server *s) {
 		free(s->instance);
 	if (s->saddr != NULL)
 		freeaddrinfo(s->saddr);
+	if (s->hint)
+		free(s->hint);
 	free((char *)s->ip);
 	s->ip = NULL;
 	free(s);
@@ -755,9 +1162,9 @@ server_instance(server *s)
 }
 
 /**
- * Returns the connection type of this server.
+ * Returns the connection protocol of this server.
  */
-inline serv_ctype
+inline con_proto
 server_ctype(server *s)
 {
 	if (s == NULL)
@@ -765,6 +1172,28 @@ server_ctype(server *s)
 	return s->ctype;
 }
 
+/**
+ * Returns the connection transport of this server.
+ */
+inline con_trnsp
+server_transport(server *s)
+{
+	if (s == NULL)
+		return W_PLAIN;
+	return s->transport;
+}
+
+/**
+ * Returns the connection type of this server.
+ */
+inline con_type
+server_type(server *s)
+{
+	if (s == NULL)
+		return T_LINEMODE;
+	return s->type;
+}
+
 /**
  * Returns whether the last action on this server caused a failure.
  */
