@@ -29,13 +29,17 @@ typedef struct _server server;
 server *server_new(
 		const char *ip,
 		unsigned short port,
-		serv_ctype ctype,
+		con_type type,
+		con_trnsp transport,
+		con_proto ctype,
 		struct addrinfo *saddr,
+		struct addrinfo *hint,
 		size_t queuesize,
 		size_t batchsize,
 		int maxstalls,
 		unsigned short iotimeout,
 		unsigned int sockbufsize);
+char server_cmp(server *s, struct addrinfo *saddr, const char *ip);
 char server_start(server *s);
 void server_add_secondaries(server *d, server **sec, size_t cnt);
 void server_set_failover(server *d);
@@ -47,7 +51,9 @@ void server_swap_queue(server *l, server *r);
 const char *server_ip(server *s);
 unsigned short server_port(server *s);
 char *server_instance(server *s);
-serv_ctype server_ctype(server *s);
+con_proto server_ctype(server *s);
+con_type server_type(server *s);
+con_trnsp server_transport(server *s);
 char server_failed(server *s);
 size_t server_get_ticks(server *s);
 size_t server_get_metrics(server *s);
